# -*- coding: utf-8 -*-
"""
Created on Sun Sep 29 18:16:09 2019

@author: qchat
"""
import configparser
import json
import datetime
import os
import math as m
from typing import Any, Tuple, List, Dict, Union
from collections import OrderedDict

import numpy as np
import pandas as pd
from qtpy import QtWidgets, QtCore

from ...config import get_scanner_config
from ...elements import Variable as Variable_og
from ...elements import Action
from ...devices import DEVICES, list_loaded_devices, get_element_by_address
from ...utilities import (boolean, str_to_array, array_to_str, create_array,
                          str_to_dataframe, dataframe_to_str, str_to_data,
                          str_to_tuple)
from ...variables import (get_variable, has_eval, is_Variable, eval_variable,
                          remove_from_config, update_from_config, VARIABLES)
from ...paths import PATHS
from .... import __version__


class ConfigHistory:
    """ Manage config history: store each change made to a config """

    def __init__(self):
        self.active = True
        self.list = []
        self.index = -1
        # Note: it's normal to have the first data unchangeable by append method
        # use pop if need to remove first data

    def __repr__(self) -> str:
        return (object.__repr__(self) + "\n\t" + self.list.__repr__()
                + f"\n\tCurrent data: {self.get_data()}")

    def __len__(self) -> int:
        return len(self.list)

    def append(self, data: dict):

        if data is not None:
            # if in middle remove right indexes
            if (self.index != len(self)-1): self.list = self.list[: self.index+1]
            self.index += 1
            self.list.append(data)

    def pop(self):  # OBSOLETE
        if len(self) != 0:
            if (self.index == len(self) - 1): self.index -= 1
            self.list.pop()

    def go_up(self):
        if (self.index + 1) <= (len(self) - 1): self.index += 1

    def go_down(self):
        if (self.index-1 >= 0): self.index -= 1

    def get_data(self) -> dict:
        return self.list[self.index] if self.index >= 0 else None


class ConfigManager:
    """ Manage a config, storing recipes of a scan """

    def __init__(self, gui: QtWidgets.QMainWindow):

        self.gui = gui

        # Import Autolab config
        scanner_config = get_scanner_config()
        self.precision = scanner_config['precision']

        # Initializing configuration values
        self.config = OrderedDict()

        self.configHistory = ConfigHistory()
        self.configHistory.active = True

        self._old_variables = []  # To update variable menu

    def ask_get_element_by_address(self, device_name: str, address: str):
        """ Wrap of :meth:`get_element_by_address` to ask user if want to
        instantiate device if not already instantiated.
        Returns the element at the given address. """
        if device_name not in DEVICES:
            msg_box = QtWidgets.QMessageBox(self.gui)
            msg_box.setWindowTitle(f"Device {device_name}")
            msg_box.setText(f"Instantiate device {device_name}?")
            msg_box.setStandardButtons(QtWidgets.QMessageBox.Ok
                                       | QtWidgets.QMessageBox.Cancel)
            msg_box.show()
            if msg_box.exec_() == QtWidgets.QMessageBox.Cancel:
                raise ValueError(f'Refused {device_name} instantiation')

        element = get_element_by_address(address)

        return element

    def update_loaded_devices(self, already_loaded_devices: list):
        """ Refresh GUI with the new loaded devices """
        for device in (set(list_loaded_devices()) - set(already_loaded_devices)):
            item_list = self.gui.mainGui.tree.findItems(
                device, QtCore.Qt.MatchExactly, 0)

            if len(item_list) == 1:
                item = item_list[0]
                self.gui.mainGui.itemClicked(item)

    # NAMES
    ###########################################################################

    def getNames(self, recipe_name: str, option: str = None) -> List[str]:
        """ Returns a list of step names and parameter names of a recipe """
        names = [step['name'] for step in self.stepList(recipe_name)]

        if option != 'recipe':
            for parameter in self.parameterList(recipe_name):
                names.append(parameter['name'])

        return names

    def getUniqueName(self, recipe_name: str, basename: str,
                      ignore: List[str] = []) -> str:
        """ Returns unique name: Adds a number next to basename in case this
        basename is already taken for the given recipe """
        name = basename
        names = []

        for recipe_name in self.getRecipeLink(recipe_name):
            if recipe_name in self.recipeNameList():
                names += self.getNames(recipe_name)

        names = [name for name in names if name not in ignore]
        compt = 0
        while True:
            if name in names:
                compt += 1
                name = basename + '_' + str(compt)
            else:
                break

        return name

    def getUniqueNameRecipe(self, basename: str, ignore: List[str] = []) -> str:
        """ Returns unique name for recipe: Adds a number next to basename in
        case this basename is already taken by a recipe """
        name = basename
        names = []

        names += self.recipeNameList()
        names = [name for name in names if name not in ignore]

        compt = 0
        while True:
            if name in names:
                compt += 1
                name = basename + '_' + str(compt)
            else:
                break

        return name

    # CONFIG MODIFICATIONS
    ###########################################################################
    def updateVariableConfig(self, new_variables: List[Tuple[str, Any]] = None):
        if new_variables is None:
            new_variables = self.getConfigVariables()
        remove_variables = list(set(self._old_variables) - set(new_variables))
        remove_from_config(remove_variables)
        update_from_config(new_variables)
        self._old_variables = new_variables

    def addNewConfig(self):
        """ Adds new config to history list """
        if self.configHistory.active:
            self.updateVariableConfig()  # before create_configPars
            self.configHistory.append(self.create_configPars())
            self.gui.undo.setEnabled(True)
            self.gui.redo.setEnabled(False)

    def addRecipe(self, recipe_name: str):
        """ Adds new recipe to config """
        if not self.gui.scanManager.isStarted():
            recipe_name = self.getUniqueNameRecipe(recipe_name)

            self.config[recipe_name] = {}
            self.config[recipe_name]['parameter'] = []
            self.config[recipe_name]['recipe'] = []
            self.config[recipe_name]['active'] = True
            self._addDefaultParameter(recipe_name)

            self.gui._addRecipe(recipe_name)
            self.addNewConfig()

    def removeRecipe(self, recipe_name: str):
        """ Removes recipe from config """
        if not self.gui.scanManager.isStarted():
            self.config.pop(recipe_name)
            self.gui._removeRecipe(recipe_name)
            self.addNewConfig()

    def activateRecipe(self, recipe_name: str, state: bool):
        """ Activates/Deactivates a recipe """
        if not self.gui.scanManager.isStarted():
            self.config[recipe_name]['active'] = bool(state)

            self.gui._activateRecipe(recipe_name, self.config[recipe_name]['active'])
            self.addNewConfig()

    def setRecipeOrder(self, keys: List[str]):
        """ Reorders recipes according to the list of recipe names 'keys' """
        if not self.gui.scanManager.isStarted():
            self.config = OrderedDict((key, self.config[key]) for key in keys)
            self.gui._resetRecipe()
            self.addNewConfig()

    def renameRecipe(self, existing_recipe_name: str, new_recipe_name: str):
        """ Renames recipe """
        if not self.gui.scanManager.isStarted():
            if new_recipe_name == existing_recipe_name:
                return None
            if existing_recipe_name not in self.recipeNameList():
                raise ValueError(
                    'should not be possible to select a non existing recipe_name: ' \
                        f'{existing_recipe_name} not in {self.recipeNameList()}')

            new_recipe_name = self.getUniqueNameRecipe(
                new_recipe_name, ignore=[existing_recipe_name])

            if new_recipe_name == existing_recipe_name:
                return None
            old_config = self.config
            new_config = {}

            for recipe_name in old_config:
                if recipe_name == existing_recipe_name:
                    new_config[new_recipe_name] = old_config[recipe_name]
                else:
                    new_config[recipe_name] = old_config[recipe_name]

            self.config = new_config

            prev_index_recipe = self.gui.selectRecipe_comboBox.currentIndex()
            prev_index_param = self.gui.selectParameter_comboBox.currentIndex()
            self.gui._resetRecipe()
            self.gui.selectRecipe_comboBox.setCurrentIndex(prev_index_recipe)
            self.gui._updateSelectParameter()
            self.gui.selectParameter_comboBox.setCurrentIndex(prev_index_param)

    def checkConfig(self):
        """ Checks validity of a config. Used before a scan start or after a scan pause. """
        assert len(self.recipeNameList()) != 0, 'Need a recipe to start a scan!'

        one_recipe_active = False
        for recipe_name in self.recipeNameList():
            recipe = self.config[recipe_name]
            if recipe['active']:
                one_recipe_active = True
                assert len(recipe['recipe']) > 0, f"Recipe {recipe_name} is empty!"

            list_recipe_new = [recipe]
            has_sub_recipe = True

            while has_sub_recipe:  # OBSOLETE
                has_sub_recipe = False
                recipe_list = list_recipe_new

                for recipe_i in recipe_list:

                    for step in recipe_i['recipe']:
                        if step['stepType'] == 'recipe':
                            has_sub_recipe = True
                            assert step['element'] in self.config, (
                                f"Recipe {step['element']} doesn't exist in {recipe_name}!")
                            other_recipe = self.config[step['element']]
                            assert len(other_recipe['recipe']) > 0, (
                                f"Recipe {step['element']} is empty!")
                            list_recipe_new.append(other_recipe)

                    list_recipe_new.remove(recipe_i)

        assert one_recipe_active, "Need at least one active recipe!"

        already_loaded_devices = list_loaded_devices()
        try:
            # Replace closed devices by reopened one
            for recipe_name in self.recipeNameList():
                for step in (self.stepList(recipe_name)
                              + self.parameterList(recipe_name)):
                    if step['element']:
                        device_name = step['element'].address().split('.')[0]
                        element = self.ask_get_element_by_address(
                            device_name, step['element'].address())
                        step['element'] = element
        finally:
            self.update_loaded_devices(already_loaded_devices)

    def lastRecipeName(self) -> str:
        """ Returns last recipe name """
        return self.recipeNameList()[-1] if len(self.recipeNameList()) != 0 else ""

    # set Param
    def _defaultParameterPars(self) -> dict:
        return {'name': 'parameter',
                'address': 'None',
                'nbpts': 1,
                'start_value': 0,
                'end_value': 0,
                'log': False}

    def _addDefaultParameter(self, recipe_name: str):
        """ Adds a default parameter to the config"""
        parameter_name = self.getUniqueName(recipe_name, 'parameter')

        parameter = {'name': parameter_name,
                     'element': None,
                     'nbpts': 11,
                     'range': (0, 10),
                     'step': 1,
                     'log': False,
                     }

        self.parameterList(recipe_name).append(parameter)

    def addParameter(self, recipe_name: str):
        """ Adds a parameter to the recipe """
        if not self.gui.scanManager.isStarted():
            self._addDefaultParameter(recipe_name)
            param_name = self.parameterList(recipe_name)[-1]['name']
            self.gui._addParameter(recipe_name, param_name)

            self.addNewConfig()

    def removeParameter(self, recipe_name: str, param_name: str):
        """ Removes the parameter from the recipe """
        if not self.gui.scanManager.isStarted():
            pos = self.getParameterPosition(recipe_name, param_name)
            self.parameterList(recipe_name).pop(pos)
            self.gui._removeParameter(recipe_name, param_name)

            self.addNewConfig()

    def setParameter(self, recipe_name: str, param_name: str,
                     element: Variable_og, newName: str = None):
        """ Sets the element provided as the new parameter of the scan.
        Add a parameter is no existing parameter """
        if not self.gui.scanManager.isStarted():
            if recipe_name == "":
                self.configHistory.active = False
                self.addRecipe("recipe")
                self.configHistory.active = True
                recipe_name = self.lastRecipeName()
                param_name = self.parameterNameList(recipe_name)[-1]
            if len(self.parameterList(recipe_name)) == 0:
                self.configHistory.active = False
                self.addParameter(recipe_name)
                self.configHistory.active = True
                param_name = self.parameterList(recipe_name)[-1]['name']

            param = self.getParameter(recipe_name, param_name)

            param['element'] = element
            if newName is None:
                if param_name == element.name:
                    newName = param_name
                else:
                    newName = self.getUniqueName(
                        recipe_name, element.name)
            param['name'] = newName
            self.gui._refreshParameterRange(recipe_name, param_name, newName)
            self.addNewConfig()

    def renameParameter(self, recipe_name: str, param_name: str, newName: str):
        """ Renames a parameter of a recipe """
        if not self.gui.scanManager.isStarted():
            if newName != param_name:
                param = self.getParameter(recipe_name, param_name)

                newName = self.getUniqueName(
                    recipe_name, newName, ignore=[param_name])
                if newName != param_name:
                    param['name'] = newName
                    self.addNewConfig()
        else:
            newName = param_name

        self.gui._refreshParameterRange(recipe_name, param_name, newName)

    def setNbPts(self, recipe_name: str, param_name: str, value: int):
        """ Sets the number of points of a parameter """
        if not self.gui.scanManager.isStarted():
            param = self.getParameter(recipe_name, param_name)

            param['nbpts'] = value

            if value == 1:
                param['step'] = 0
            else:
                xrange = tuple(self.getRange(recipe_name, param_name))
                width = abs(xrange[1] - xrange[0])
                param['step'] = width / (value - 1)
            self.addNewConfig()

        self.gui._refreshParameterRange(recipe_name, param_name)

    def setStep(self, recipe_name: str, param_name: str, value: float):
        """ Sets the step between points of a parameter """
        if not self.gui.scanManager.isStarted():
            param = self.getParameter(recipe_name, param_name)

            if value == 0:
                param['nbpts'] = 1
                param['step'] = 0
            else:
                param['step'] = value
                xrange = tuple(self.getRange(recipe_name, param_name))
                width = abs(xrange[1] - xrange[0])
                param['nbpts'] = int(round(width / value) + 1)
                if width != 0:
                    param['step'] = width / (param['nbpts'] - 1)
            self.addNewConfig()

        self.gui._refreshParameterRange(recipe_name, param_name)

    def setRange(self, recipe_name: str, param_name: str,
                 lim: Tuple[float, float]):
        """ Sets the range values (start and end value) of a parameter """
        if not self.gui.scanManager.isStarted():
            param = self.getParameter(recipe_name, param_name)

            if lim != param['range']: param['range'] = tuple(lim)

            width = abs(lim[1] - lim[0])

            recipeDictRange = self.gui.recipeDict[recipe_name]['parameterManager']

            if recipeDictRange[param_name].point_or_step == "point":
                param['step'] = float(
                    width / (self.getNbPts(recipe_name, param_name) - 1))

            elif recipeDictRange[param_name].point_or_step == "step":
                param['nbpts'] = int(
                    round(width / self.getStep(recipe_name, param_name)) + 1)
                param['step'] = float(
                    width / (self.getNbPts(recipe_name, param_name) - 1))
            self.addNewConfig()

        self.gui._refreshParameterRange(recipe_name, param_name)

    def setLog(self, recipe_name: str, param_name: str, state: bool):
        """ Sets the log state of a parameter """
        if not self.gui.scanManager.isStarted():
            param = self.getParameter(recipe_name, param_name)

            if state != param['log']: param['log'] = state
            self.addNewConfig()

        self.gui._refreshParameterRange(recipe_name, param_name)

    def setValues(self, recipe_name: str, param_name: str, values: List[float]):
        """ Sets custom values to a parameter """
        if not self.gui.scanManager.isStarted():
            param = self.getParameter(recipe_name, param_name)

            if has_eval(values) or np.ndim(values) == 1:
                param['values'] = values
                self.addNewConfig()

        self.gui._refreshParameterRange(recipe_name, param_name)

    # set step
    def addRecipeStep(self, recipe_name: str, stepType: str, element,
                      name: str = None, value = None):
        """ Adds a step to the scan recipe """
        if recipe_name == "":
            self.addRecipe("recipe")
            recipe_name = self.lastRecipeName()

        if not self.gui.scanManager.isStarted():
            assert recipe_name in self.recipeNameList(), (
                f'{recipe_name} not in {self.recipeNameList()}')

            if name is None:
                name = self.getUniqueName(recipe_name, element.name)
            else:
                name = self.getUniqueName(recipe_name, name)

            step = {'name': name, 'element': element,
                    'stepType': stepType, 'value': None}

            # Value
            if stepType == 'recipe':
                assert element != recipe_name, "Can't have a recipe in itself: {element}"  # safeguard but should be stopped before arriving here
            if stepType == 'set': setValue = True
            elif stepType == 'action' and element.type in [
                    int, float, bool, str, bytes, tuple, np.ndarray, pd.DataFrame]:
                setValue = True
            else: setValue = False

            if setValue:
                if value is None:
                    if element.type in [int, float]: value = 0
                    elif element.type in [str]: value = ''
                    elif element.type in [bytes]: value = b''
                    elif element.type in [pd.DataFrame]: value = pd.DataFrame()
                    elif element.type in [np.ndarray]: value = np.array([])
                    elif element.type in [bool]: value = False
                    elif element.type in [tuple]: value = ([], -1)
                step['value'] = value

            self.stepList(recipe_name).append(step)
            self.gui._refreshRecipe(recipe_name)
            self.addNewConfig()

    def delRecipeStep(self, recipe_name: str, name: str):
        """ Removes a step from the scan recipe """
        if not self.gui.scanManager.isStarted():
            pos = self.getRecipeStepPosition(recipe_name, name)
            self.stepList(recipe_name).pop(pos)
            self.gui._refreshRecipe(recipe_name)
            self.addNewConfig()

    def renameRecipeStep(self, recipe_name: str, name: str, newName: str):
        """ Renames a step in the scan recipe """
        if not self.gui.scanManager.isStarted():
            if newName != name:
                step_info = self.getRecipeStep(recipe_name, name)
                newName = self.getUniqueName(recipe_name, newName)
                step_info['name'] = newName
                self.gui._refreshRecipe(recipe_name)
                self.addNewConfig()

    def setRecipeStepValue(self, recipe_name: str, name: str, value: Any):
        """ Sets the value of a step in the scan recipe """
        if not self.gui.scanManager.isStarted():
            step_info = self.getRecipeStep(recipe_name, name)
            if value is not step_info['value']:
                step_info['value'] = value
                self.gui._refreshRecipe(recipe_name)
                self.addNewConfig()

    def setRecipeStepOrder(self, recipe_name: str, stepOrder: list):
        """ Reorders steps of a recipe according to the list of step names 'stepOrder' """
        if not self.gui.scanManager.isStarted():
            newOrder = [self.getRecipeStepPosition(
                recipe_name, name) for name in stepOrder]
            recipe = self.stepList(recipe_name)
            self.config[recipe_name]['recipe'] = [recipe[i] for i in newOrder]
            self.addNewConfig()

        self.gui._refreshRecipe(recipe_name)

    # CONFIG READING
    ###########################################################################
    def recipeNameList(self):
        """ Returns the list of recipe names """
        return list(self.config.keys())

    def getLinkedRecipe(self) -> Dict[str, list]:  # OBSOLETE
        """ Returns a dict with recipe_name key and list of recipes linked to recipe_name recipe.
        Example: {'recipe_1': ['recipe_1', 'recipe_2', 'recipe_3', 'recipe_2'], 'recipe_3': ['recipe_3', 'recipe_2'], 'recipe_2': ['recipe_2']}"""
        linkedRecipe = {}

        for recipe_name in self.recipeNameList():
            recipe = self.config[recipe_name]
            recipe_name_link = []
            list_recipe_new = [recipe]
            has_sub_recipe = True

            i = 0
            while has_sub_recipe:
                i +=1
                if i > 100:  # safeguard, condition: "step['element'] in recipe_name_link" should be enough
                    raise ValueError('ERROR: stopped recursive function to prevent crash')
                has_sub_recipe = False
                recipe_list = list_recipe_new
                for recipe_i in recipe_list:

                    for step in recipe_i['recipe']:
                        if step['stepType'] == 'recipe':
                            if step['element'] in recipe_name_link:
                                raise ValueError(f"ERROR: cycle detected for recipe {step['element']}!")
                            has_sub_recipe = True
                            if step['element'] in self.config:
                                other_recipe = self.config[step['element']]
                                list_recipe_new.append(other_recipe)
                                recipe_name_link.append(step['element'])

                    list_recipe_new.remove(recipe_i)

            recipe_name_link.append(recipe_name)
            linkedRecipe[recipe_name] = recipe_name_link

        return linkedRecipe

    def getRecipeLink(self, recipe_name: str) -> List[str]:  # OBSOLETE
        """ Returns a list of unique recipe names for which recipes are linked to recipe_name
        Example: for 'recipe_1': ['recipe_1', 'recipe_2', 'recipe_3'] """
        linkedRecipe = self.getLinkedRecipe()
        uniqueLinkedRecipe = []

        for key, val in linkedRecipe.items():
            if recipe_name in val:
                uniqueLinkedRecipe.append(val)

        return list(set(sum(uniqueLinkedRecipe, [])))

    def getAllowedRecipe(self, recipe_name: str) -> List[str]:  # OBSOLETE
        """ Returns a list of recipe that can be added to recipe_name without
        risk of cycle or twice same recipe """
        recipe_name_list = self.recipeNameList()
        linked_recipes = self.getLinkedRecipe()

        for recipe_name_i, recipe_i in linked_recipes.items():
            # remove recipe that are in recipe_name
            if recipe_name_i in linked_recipes[recipe_name]:
                if recipe_name_i in recipe_name_list:
                    recipe_name_list.remove(recipe_name_i)

            # remove recipe that contains recipe_name
            if recipe_name in recipe_i:
                if recipe_name_i in recipe_name_list:
                    recipe_name_list.remove(recipe_name_i)

                # remove all recipes that are in recipe_name_i
                for recipe_name_j in recipe_i:
                    if recipe_name_j in recipe_name_list:
                        recipe_name_list.remove(recipe_name_j)

        return recipe_name_list

    def getActive(self, recipe_name: str) -> bool:
        """ Returns whether the recipe is active """
        return self.config[recipe_name]['active']

    def getRecipeActive(self) -> List[str]:
        """ Returns list of active recipes """
        return [i for i in self.recipeNameList() if self.getActive(i)]

    # get Param
    def getParameter(self, recipe_name: str, param_name: str) -> dict:
        pos = self.getParameterPosition(recipe_name, param_name)
        return self.parameterList(recipe_name)[pos]

    def getParameterPosition(self, recipe_name: str, param_name: str) -> int:
        """ Returns the position of a parameter """
        return [i for i, param in enumerate(self.parameterList(recipe_name))
                if param['name'] == param_name][0]

    def parameterList(self, recipe_name: str) -> List[dict]:
        """ Returns the list of parameters in the recipe """
        return self.config[recipe_name]['parameter']

    def parameterNameList(self, recipe_name: str) -> List[str]:
        """ Returns the list of parameter names in the recipe """
        return [param['name'] for param in self.parameterList(recipe_name)] if (
            recipe_name in self.config) else []

    def getParameterElement(
            self, recipe_name: str, param_name: str) -> Union[None, Variable_og]:
        """ Returns the element of a parameter """
        param = self.getParameter(recipe_name, param_name)
        return param['element']

    def getLog(self, recipe_name: str, param_name: str) -> bool:
        """ Returns the log state of a parameter """
        param = self.getParameter(recipe_name, param_name)
        return param['log']

    def getNbPts(self, recipe_name: str, param_name: str) -> int:
        """ Returns the number of points of a parameter """
        param = self.getParameter(recipe_name, param_name)
        return param['nbpts']

    def getStep(self, recipe_name: str, param_name: str) -> float:
        """ Returns the step between points of a parameter """
        param = self.getParameter(recipe_name, param_name)
        return param['step']

    def getRange(self, recipe_name: str, param_name: str) -> Tuple[float, float]:
        """ Returns the range (start and end value) of a parameter """
        param = self.getParameter(recipe_name, param_name)
        return param['range']

    def getValues(self, recipe_name: str, param_name: str) -> List[float]:
        """ Returns the values of a parameter """
        param = self.getParameter(recipe_name, param_name)
        if 'values' in param:
            return param['values']
        else:
            startValue, endValue = param['range']
            nbpts = param['nbpts']
            logScale = param['log']

            # Creates the array of values for the parameter
            if logScale:
                paramValues = np.logspace(
                    m.log10(startValue), m.log10(endValue), nbpts, endpoint=True)
            else:
                paramValues = np.linspace(startValue, endValue, nbpts, endpoint=True)

            return paramValues

    def hasCustomValues(self, recipe_name: str, param_name: str) -> bool:
        """ Boolean to know if parameter has custom array """
        param = self.getParameter(recipe_name, param_name)
        return 'values' in param

    # get Step
    def stepList(self, recipe_name: str) -> List[dict]:
        """ Returns the list of steps in the recipe """
        return self.config[recipe_name]['recipe']

    def getRecipeStep(self, recipe_name: str, name: str) -> dict:
        """ Returns a dictionnary with recipe step information """
        pos = self.getRecipeStepPosition(recipe_name, name)
        return self.stepList(recipe_name)[pos]

    def getRecipeStepElement(
            self, recipe_name: str, name: str) -> Union[Variable_og, Action]:
        """ Returns the element of a recipe step """
        step_info = self.getRecipeStep(recipe_name, name)
        return step_info['element']

    def getRecipeStepType(self, recipe_name: str, name: str) -> str:
        """ Returns the type a recipe step """
        step_info = self.getRecipeStep(recipe_name, name)
        return step_info['stepType']

    def getRecipeStepValue(self, recipe_name: str, name: str) -> Any:
        """ Returns the value of a recipe step """
        step_info = self.getRecipeStep(recipe_name, name)
        return step_info['value']

    def getRecipeStepPosition(self, recipe_name: str, name: str) -> int:
        """ Returns the position of a recipe step in the recipe """
        return [i for i, step in enumerate(self.stepList(recipe_name)) if step['name'] == name][0]

    def getParamDataFrame(self, recipe_name: str, param_name: str) -> pd.DataFrame:
        """ Returns a pd.DataFrame with 'id' and 'param_name'
        columns containing the parameter array """
        paramValues = self.getValues(recipe_name, param_name)
        paramValues = eval_variable(paramValues)
        paramValues = create_array(paramValues)
        assert isinstance(paramValues, np.ndarray)
        data = pd.DataFrame()
        data["id"] = 1 + np.arange(len(paramValues))
        data[param_name] = paramValues

        return data

    def getConfigVariables(self) -> List[Tuple[str, Any]]:
        """ Returns a (key, value) list of parameters and measured step """
        listVariable = []
        listVariable.append(('ID', 1))

        for recipe_name in reversed(self.recipeNameList()):
            for param_name in self.parameterNameList(recipe_name):
                values = self.getValues(recipe_name, param_name)
                value = values if has_eval(values) else float(values[0])
                listVariable.append((param_name, value))
            for step in self.stepList(recipe_name):
                if step['stepType'] == 'measure':
                    listVariable.append((step['name'], step['value']))

        return listVariable

    # EXPORT IMPORT ACTIONS
    ###########################################################################

    def export(self, filename: str):
        """ Exports the current scan configuration in the provided file """
        configPars = self.create_configPars()

        with open(filename, "w") as configfile:
            json.dump(configPars, configfile, indent=4)

    def create_configPars(self) -> dict:
        """ Creates the current scan configuration parser """
        configPars = {}
        configPars['autolab'] = {'version': __version__,
                                 'timestamp': str(datetime.datetime.now())}

        for recipe_num, recipe_name in enumerate(self.recipeNameList()):
            recipe_i = self.config[recipe_name]
            pars_recipe_i = {}

            pars_recipe_i['name'] = str(recipe_name)
            pars_recipe_i['active'] = str(bool(recipe_i['active']))
            pars_recipe_i['parameter'] = {}

            for i, param in enumerate(recipe_i['parameter']):
                param_name = f"parameter_{i+1}"

                param_pars = {}

                param_pars['name'] = param['name']

                if param['element'] is not None:
                    param_pars['address'] = param['element'].address()
                else:
                    param_pars['address'] = "None"

                if 'values' in param:
                    if has_eval(param['values']):
                        param_pars['values'] = param['values']
                    else:
                        param_pars['values'] = array_to_str(
                            param['values'], threshold=1000000, max_line_width=9000000)
                else:
                    param_pars['nbpts'] = str(param['nbpts'])
                    param_pars['start_value'] = str(param['range'][0])
                    param_pars['end_value'] = str(param['range'][1])
                    param_pars['log'] = str(int(param['log']))

                pars_recipe_i['parameter'][param_name] = param_pars

            pars_recipe_i['recipe'] = {}

            for i, config_step in enumerate(recipe_i['recipe']):
                pars_recipe_i['recipe'][f'{i+1}_name'] = config_step['name']
                pars_recipe_i['recipe'][f'{i+1}_steptype'] = config_step['stepType']

                if config_step['stepType'] == 'recipe':
                    pars_recipe_i['recipe'][f'{i+1}_address'] = config_step['element']
                else:
                    pars_recipe_i['recipe'][f'{i+1}_address'] = config_step['element'].address()

                stepType = config_step['stepType']

                if stepType == 'set' or (stepType == 'action'
                                         and config_step['element'].type in [
                                             int, float, bool, str, bytes, tuple,
                                             np.ndarray, pd.DataFrame]):
                    value = config_step['value']

                    if has_eval(value):
                        valueStr = value
                    else:
                        if config_step['element'].type in [np.ndarray]:
                            valueStr = array_to_str(
                                value, threshold=1000000, max_line_width=9000000)
                        elif config_step['element'].type in [pd.DataFrame]:
                            valueStr = dataframe_to_str(value, threshold=1000000)
                        elif config_step['element'].type in [int, float, str]:
                            try:
                                valueStr = f'{value:.{self.precision}g}'
                            except:
                                valueStr = f'{value}'
                        elif config_step['element'].type in [bytes]:
                            valueStr = f'{value.decode()}'
                        else:  # for tuple and safety
                            valueStr = f'{value}'

                    pars_recipe_i['recipe'][f'{i+1}_value'] = valueStr

            configPars[f"recipe_{recipe_num+1}"] = pars_recipe_i

        # Add variables to config
        name_var_config = [var[0] for var in self.getConfigVariables()]
        names_var_user = list(VARIABLES)
        names_var_to_save = list(set(names_var_user) - set(name_var_config))

        var_to_save = {}
        for var_name in names_var_to_save:
            var = get_variable(var_name)

            if var is not None:
                assert is_Variable(var)
                value_raw = var.raw
                if isinstance(value_raw, np.ndarray):
                    valueStr = array_to_str(
                        value_raw, threshold=1000000, max_line_width=9000000)
                elif isinstance(value_raw, pd.DataFrame):
                    valueStr = dataframe_to_str(value_raw, threshold=1000000)
                elif isinstance(value_raw, (int, float, str)):
                    try: valueStr = f'{value_raw:.{self.precision}g}'
                    except: valueStr = f'{value_raw}'

                var_to_save[var_name] = valueStr

        if len(var_to_save) != 0: configPars['variables'] = var_to_save

        return configPars

    def import_configPars(self, filename: str, append: bool = False):
        """ Import a scan configuration from file with filename name """
        if not self.gui.scanManager.isStarted():
            if os.path.exists(filename):
                try:
                    legacy_configPars = configparser.ConfigParser()
                    legacy_configPars.read(filename)
                except:
                    try:
                        with open(filename, "r") as read_file:
                            configPars = json.load(read_file)
                    except Exception as e:
                        self.gui.setStatus(
                            f"Impossible to load configuration file: {e}",
                            10000, False)
                        return None
                else:
                    print("ConfigParser depreciated, now use json. " \
                          "Will convert this config to json if save it.")
                    configPars = {s: dict(legacy_configPars.items(s))
                                  for s in legacy_configPars.sections()}

                path = os.path.dirname(filename)
                PATHS['last_folder'] = path

                self.load_configPars(configPars, append=append)

                if not self._got_error: self.addNewConfig()
            else:
                self.gui.setStatus(
                    f"Configuration file {filename} doesn't exists", 5000)

    def load_configPars(self, configPars: dict, append: bool = False):
        """ Creates a config representing a scan form a configPars """
        self._got_error = False
        self.configHistory.active = False
        previous_config = self.config.copy()  # used to recover old config if error in loading new one
        already_loaded_devices = list_loaded_devices()

        try:
            # LEGACY <= 1.2
            try:
                LEGACY = (configPars["autolab"]["version"].startswith("1.0")
                          or configPars["autolab"]["version"].startswith("1.1.")
                          or configPars["autolab"]["version"] == "1.2")
                if LEGACY:
                    new_configPars = OrderedDict()
                    new_configPars["autolab"] = configPars["autolab"]

                    if 'initrecipe' in configPars:
                        new_configPars["recipe_1"] = {}
                        new_configPars["recipe_1"]['name'] = "init"
                        new_configPars["recipe_1"]['active'] = "True"
                        new_configPars["recipe_1"]['parameter'] = {}
                        new_configPars["recipe_1"]['parameter']["parameter_1"] = self._defaultParameterPars()
                        new_configPars["recipe_1"]['parameter']["parameter_1"]['name'] = 'init'
                        new_configPars["recipe_1"]['recipe'] = configPars['initrecipe']

                    new_configPars["recipe_2"] = {}
                    new_configPars["recipe_2"]['name'] = "recipe_1"
                    new_configPars["recipe_2"]['active'] = "True"
                    new_configPars["recipe_2"]['parameter'] = {}
                    new_configPars["recipe_2"]['parameter']["parameter_1"] = configPars['parameter']
                    new_configPars["recipe_2"]['recipe'] = configPars['recipe']

                    if 'endrecipe' in configPars:
                        new_configPars["recipe_3"] = {}
                        new_configPars["recipe_3"]['name'] = "end"
                        new_configPars["recipe_3"]['active'] = "True"
                        new_configPars["recipe_3"]['parameter'] = {}
                        new_configPars["recipe_3"]['parameter']["parameter_1"] = self._defaultParameterPars()
                        new_configPars["recipe_3"]['parameter']['name'] = 'end'
                        new_configPars["recipe_3"]['recipe'] = configPars['endrecipe']

                    configPars = new_configPars
            except: pass

            # Config
            config = OrderedDict()
            # to remove 'autolab' and 'variables' from recipe list
            recipeNameList = [i for i in list(configPars)
                              if i not in ('autolab', 'variables')]

            for recipe_num_name in recipeNameList:

                pars_recipe_i = configPars[recipe_num_name]

                if 'name' in pars_recipe_i:
                    recipe_name = pars_recipe_i['name']
                else:
                    recipe_name = recipe_num_name  # LEGACY <= 2.0b1

                config[recipe_name] = OrderedDict()
                recipe_i = config[recipe_name]

                recipe_i['name'] = recipe_name

                if 'active' in pars_recipe_i:
                    recipe_i['active'] = boolean(pars_recipe_i['active'])
                else:
                    recipe_i['active'] = True  # LEGACY <= 1.2.1

                assert 'parameter' in pars_recipe_i, (
                    f'Missing parameter in {recipe_name}')

                param_list = recipe_i['parameter'] = []

                # LEGACY <= 1.2.1
                if len(pars_recipe_i['parameter']) != 0:
                    if not isinstance(
                            list(pars_recipe_i['parameter'].values())[0],
                            dict):
                        pars_recipe_i['parameter'] = {
                            'parameter_1': pars_recipe_i['parameter']}

                for param_pars_name in pars_recipe_i['parameter']:
                    param_pars = pars_recipe_i['parameter'][param_pars_name]

                    param = {}

                    assert 'name' in param_pars, (
                        f"Missing name to {param_pars}")
                    param['name'] = param_pars['name']

                    assert 'address' in param_pars, (
                        f"Missing address to {param_pars}")
                    if param_pars['address'] == "None": element = None
                    else:
                        device_name = param_pars['address'].split('.')[0]
                        element = self.ask_get_element_by_address(device_name, param_pars['address'])

                    param['element'] = element

                    if 'values' in param_pars:
                        if has_eval(param_pars['values']):
                            values = param_pars['values']
                        else:
                            values = str_to_array(param_pars['values'])
                        if not has_eval(values):
                            assert np.ndim(values) == 1, (
                                f"Values must be one dimension array in parameter: {param['name']}")
                        param['values'] = values
                    else:
                        for key in ['nbpts', 'start_value', 'end_value', 'log']:
                            assert key in param_pars, "Missing parameter key {key}."

                        param['nbpts'] = int(param_pars['nbpts'])
                        start = float(param_pars['start_value'])
                        end = float(param_pars['end_value'])
                        param['range'] = (start, end)
                        param['log'] = bool(int(param_pars['log']))

                        if param['nbpts'] > 1:
                            param['step'] = abs(end - start) / (param['nbpts'] - 1)
                        else:
                            param['step'] = 0

                    param_list.append(param)

                recipe_i['recipe'] = []
                recipe = recipe_i['recipe']
                pars_recipe = pars_recipe_i['recipe']

                while True:
                    step = {}
                    i = len(recipe) + 1

                    if f'{i}_name' in pars_recipe:
                        step['name'] = pars_recipe[f'{i}_name']
                        name = step['name']

                        assert f'{i}_steptype' in pars_recipe, (
                            f"Missing stepType in step {i} ({name}).")
                        step['stepType'] = pars_recipe[f'{i}_steptype']

                        assert f'{i}_address' in pars_recipe, (
                            f"Missing address in step {i} ({name}).")
                        address = pars_recipe[f'{i}_address']

                        if step['stepType'] == 'recipe':
                            assert step['stepType'] != 'recipe', (
                                "Removed the recipe in recipe feature!")
                            element = address
                        else:
                            device_name = address.split('.')[0]
                            element = self.ask_get_element_by_address(device_name, address)

                        step['element'] = element

                        if (step['stepType'] == 'set') or (
                                step['stepType'] == 'action' and element.type in [
                                    int, float, bool, str, bytes, tuple,
                                    np.ndarray, pd.DataFrame]):
                            assert f'{i}_value' in pars_recipe, (
                                f"Missing value in step {i} ({name}).")
                            value = pars_recipe[f'{i}_value']

                            try:
                                try:
                                    assert has_eval(value), (
                                        "Need $eval: to evaluate the given string")
                                except:
                                    # Type conversions
                                    if element.type in [int]:
                                        value = int(value)
                                    elif element.type in [float]:
                                        value = float(value)
                                    elif element.type in [str]:
                                        value = str(value)
                                    elif element.type in [bytes]:
                                        value = value.encode()
                                    elif element.type in [bool]:
                                        value = boolean(value)
                                    elif element.type in [tuple]:
                                        value = str_to_tuple(value)
                                    elif element.type in [np.ndarray]:
                                        value = str_to_array(value)
                                    elif element.type in [pd.DataFrame]:
                                        value = str_to_dataframe(value)
                                    else:
                                        assert has_eval(value), (
                                            "Need $eval: to evaluate the given string")
                            except:
                                raise ValueError(f"Error with {i}_value = {value}. Expect either {element.type} or device address. Check address or open device first.")

                            step['value'] = value
                        else:
                            step['value'] = None

                        recipe.append(step)
                    else:
                        break
            if append:
                for conf in config.values():
                    recipe_name = conf['name']
                    if recipe_name in self.recipeNameList():
                        new_recipe_name = self.getUniqueNameRecipe('recipe')
                    else:
                        new_recipe_name = recipe_name

                    self.config[new_recipe_name] = conf
            else:
                self.config = config

            if 'variables' in configPars:
                var_dict = configPars['variables']

                add_vars = []
                for var_name, raw_value in var_dict.items():
                    if not has_eval(raw_value):
                        raw_value = str_to_data(raw_value)
                    add_vars.append((var_name, raw_value))

                update_from_config(add_vars)

        except Exception as error:
            self._got_error = True
            self.gui.setStatus(
                f"Impossible to load configuration file: {error}", 10000, False)
            self.config = previous_config
        else:
            self.gui._resetRecipe()
            self.gui.setStatus("Configuration file loaded successfully", 5000)
        finally:
            self.configHistory.active = True
            self.update_loaded_devices(already_loaded_devices)

    # UNDO REDO ACTIONS
    ###########################################################################

    def undoClicked(self):
        """ Undos an action from parameter, recipe or range """
        if not self.gui.scanManager.isStarted():
            self.configHistory.go_down()
            self.changeConfig()

    def redoClicked(self):
        """ Redos an action from parameter, recipe or range """
        if not self.gui.scanManager.isStarted():
            self.configHistory.go_up()
            self.changeConfig()

    def changeConfig(self):
        """ Gets config from history and enables/disables undo/redo button accordingly """
        configPars = self.configHistory.get_data()

        if configPars is not None:
            self.load_configPars(configPars)

        self.updateUndoRedoButtons()
        self.updateVariableConfig()
        self.gui.setStatus('')

    def updateUndoRedoButtons(self):
        """ enables/disables undo/redo button depending on history """
        if self.configHistory.index == 0:
            self.gui.undo.setEnabled(False)
        else:
            self.gui.undo.setEnabled(True)

        if self.configHistory.index == len(self.configHistory)-1:
            self.gui.redo.setEnabled(False)
        else:  # implicit <
            self.gui.redo.setEnabled(True)



# -*- coding: utf-8 -*-
"""
Python package for scientific experiments automation

The purpose of this package it to provide easy and efficient tools to deal with your scientific instruments, and to run automated experiments with them, by command line instructions or through a graphical user interface (GUI).

Created by Quentin Chateiller, Python drivers originally from Quentin Chateiller and Bruno Garbin, for the C2N-CNRS (Center for Nanosciences and Nanotechnologies, Palaiseau, France) ToniQ team.
Project continued by Jonathan Peltier, for the C2N-CNRS Minaphot team and Mathieu Jeannin, for the C2N-CNRS Odin team.

Project hosted at https://github.com/autolab-project/autolab

Drivers made for this package can be found at https://github.com/autolab-project/autolab-drivers

Visit https://autolab.readthedocs.io/ for the full documentation of this package.
"""
import numpy  # OPTIMIZE: temporary fix to an infinite loading on some computer following the master merge (commit 25fd4d6)
import socket  # OPTIMIZE: temporary fix to an infinite loading on some computer

# Load current version in version file
from .core.paths import PATHS, DRIVER_SOURCES
with open(PATHS['version']) as version_file:
    __version__ = version_file.read().strip()
del version_file

# Process updates from previous versions
from .core import version_adapter
version_adapter.process_all_changes()
del version_adapter

# Load user config
from .core import config as _config
FIRST = _config.initialize_local_directory()
_config.check_autolab_config()
_config.check_plotter_config()
_config.set_temp_folder()
_config.add_extra_driver_path()
_config.add_extra_driver_repo_url()

# Add drivers folder to sys (allows a driver to import another driver)
import sys
# Order of append between local and official matter for priority
for folder in reversed(list(DRIVER_SOURCES.values())):
    sys.path.append(folder)
del sys
del folder

# infos
from .core.infos import infos, config_help
from .core.infos import _list_devices as list_devices
from .core.infos import _list_drivers as list_drivers

# Devices
from .core.devices import get_device, close, list_loaded_devices
from .core import devices as _devices

# Drivers
from .core.drivers import get_driver, explore_driver
from .core import drivers as _drivers

# Webbrowser shortcuts
from .core.web import report, doc

# Server
from .core.server import Server as server

# GUI
from .core.gui import (gui, plotter, monitor, slider, add_device, about,
                       variables_menu, preferences, driver_installer)

from .core.variables import get_variable, list_variables
from .core.variables import set_variable as add_variable

# Repository
from .core.repository import install_drivers
from .core import repository as _repository
_repository._check_empty_driver_folder()

del numpy
del socket

from .core._create_shortcut import create_shortcut

if FIRST:
    # Ask if create shortcut
    create_shortcut(ask=True)
del FIRST

# Loading the drivers informations on startup
_drivers.update_drivers_paths()

# Used by os shell to start autolab
from ._entry_script import main as _main



# -*- coding: utf-8 -*-
"""
Created on Tue Oct  1 17:38:15 2019

@author: qchat
"""
import os
import sys
import inspect
import importlib
from typing import Type, List, Tuple
from types import ModuleType

from .paths import PATHS, DRIVERS_PATHS, DRIVER_SOURCES


# =============================================================================
# DRIVERS INSTANTIATION
# =============================================================================

def get_driver(driver_name: str, connection: str, **kwargs) -> Type:
    ''' Returns a driver instance using configuration provided in kwargs '''
    if driver_name == 'autolab_server':
        from .server import Driver_REMOTE  # avoid circular import
        driver_instance = Driver_REMOTE(**kwargs)
    else:
        assert driver_name in list_drivers(), f"Driver {driver_name} not found in autolab's drivers"
        driver_lib = load_driver_lib(driver_name)
        # Need to add the driver path to allow driver imports from its folder (and only his own, not other drivers)
        driver_path = os.path.dirname(driver_lib.__file__)
        if driver_path not in sys.path:
            sys.path.append(driver_path)
        try:
            driver_instance = get_connection_class(driver_lib, connection)(**kwargs)
        finally:
            sys.path.remove(driver_path)

    return driver_instance


def load_driver_lib(driver_name: str) -> ModuleType:
    ''' Returns a driver library that contains Driver, Driver_XXX, Module_XXX '''
    # Loading preparation
    driver_path = get_driver_path(driver_name)
    driver_directory = os.path.dirname(driver_path)

    if driver_directory not in sys.path:
        sys.path.append(driver_directory)
    try:
        # Load library
        driver_lib = load_lib(driver_path)
    finally:
        sys.path.remove(driver_directory)

    return driver_lib


def load_lib(lib_path: str) -> ModuleType:
    ''' Returns an instance of the python script located at lib_path '''
    lib_name = os.path.basename(lib_path).split('.')[0]

    # Save current working directory path
    curr_dir = os.getcwd()

    # Go to the driver's directory (in case it contains absolute imports)
    os.chdir(os.path.dirname(lib_path))

    try:
        # Load the module
        spec = importlib.util.spec_from_file_location(lib_name, lib_path)
        lib = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(lib)
    finally:
        # Come back to previous working directory
        os.chdir(curr_dir)

    return lib


def load_driver_utilities_lib(driver_utilities_name: str) -> ModuleType:
    ''' Returns a driver library that contains Driver, Driver_XXX, Module_XXX '''
    # Loading preparation
    if os.path.exists(driver_utilities_name):
        driver_path = get_driver_path(driver_utilities_name.replace('_utilities', ''))
    else:
        driver_path = os.path.join(PATHS['autolab_folder'], 'core', 'default_driver.py')

    # Load library
    driver_lib = load_utilities_lib(driver_path)

    return driver_lib


def load_utilities_lib(lib_path: str) -> ModuleType:
    ''' Returns an instance of the python script located at lib_path '''
    lib_name = os.path.basename(lib_path).split('.')[0]

    # Save current working directory path
    curr_dir = os.getcwd()

    # Go to the driver's directory (in case it contains absolute imports)
    os.chdir(os.path.dirname(lib_path))

    # Load the module
    lib_name = lib_name + '_utilities'
    spec = importlib.util.spec_from_file_location(
        lib_name, os.path.join(os.path.dirname(lib_path), f'{lib_name}.py'))
    lib = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(lib)

    # Come back to previous working directory
    os.chdir(curr_dir)

    return lib


# =============================================================================
# DRIVERS LIST HELP
# =============================================================================

def list_drivers() -> List[str]:
    ''' Returns the list of available drivers '''
    # To be sure that the list is up to date
    update_drivers_paths()
    return sorted(list(DRIVERS_PATHS))


# =============================================================================
# DRIVERS INSPECTION
# =============================================================================

def get_module_names(driver_lib: ModuleType) -> List[str]:
    ''' Returns the list of the driver's Module(s) name(s) (classes Module_XXX) '''
    return [name.split('_')[1]
            for name, obj in inspect.getmembers(driver_lib, inspect.isclass)
            if obj.__module__ is driver_lib.__name__
            and name.startswith('Module_')]


def get_connection_names(driver_lib: ModuleType) -> List[str]:
    ''' Returns the list of the driver's connection types (classes Driver_XXX) '''
    return [name.split('_')[1]
            for name, obj in inspect.getmembers(driver_lib, inspect.isclass)
            if obj.__module__ is driver_lib.__name__
            and name.startswith('Driver_')]


def get_driver_category(driver_name: str) -> str:
    ''' Returns the driver's category from class Driver '''
    for filename in ('', '_utilities'):

        driver_utilities_path = os.path.join(
            os.path.dirname(get_driver_path(driver_name)), f'{driver_name}{filename}.py')
        category = 'Unknown'

        if os.path.exists(driver_utilities_path):
            try:
                driver_utilities = load_lib(driver_utilities_path)
            except Exception as e:
                print(f"Can't load {driver_name}: {e}", file=sys.stderr)
            else:
                if hasattr(driver_utilities, 'category'):
                    category = driver_utilities.category
                    break

    return category


def get_driver_class(driver_lib: ModuleType) -> Type:
    ''' Returns the class Driver of the provided driver library '''
    assert hasattr(driver_lib, 'Driver'), f"Class Driver missing in driver {driver_lib.__name__}"
    assert inspect.isclass(driver_lib.Driver), f"Class Driver missing in driver {driver_lib.__name__}"
    return driver_lib.Driver


def get_connection_class(driver_lib: ModuleType, connection: str) -> Type:
    ''' Returns the class Driver_XXX of the provided driver library and connection type '''
    if connection in get_connection_names(driver_lib):
        return getattr(driver_lib, f'Driver_{connection}')

    driver_instance = create_default_driver_conn(driver_lib, connection)
    if driver_instance is not None:
        print(f'Warning, connection {connection} not find in driver {driver_lib.__name__} but will try to connect using default connection')
        return driver_instance

    assert connection in get_connection_names(driver_lib), f"Invalid connection type {connection} for driver {driver_lib.__name__}. Try using one of this connections: {get_connection_names(driver_lib)}"


def create_default_driver_conn(driver_lib: ModuleType, connection: str) -> Type:
    """ Create a default connection class when not provided in Driver.
    Will be used to try to connect to an instrument. """
    Driver = getattr(driver_lib, 'Driver')

    if connection == 'DEFAULT':
        class Driver_DEFAULT(Driver):
            def __init__(self):
                Driver.__init__(self)

        return Driver_DEFAULT

    if connection == 'VISA':
        class Driver_VISA(Driver):
            def __init__(self, address: str = 'GPIB0::2::INSTR',
                         **kwargs):
                import pyvisa as visa

                self.TIMEOUT = 15000  # ms

                rm = visa.ResourceManager()
                self.controller = rm.open_resource(address)
                self.controller.timeout = self.TIMEOUT

                Driver.__init__(self)

            def close(self):
                try: self.controller.close()
                except: pass

            def query(self, command: str) -> str:
                result = self.controller.query(command)
                result = result.strip('\n')
                return result

            def write(self, command: str):
                self.controller.write(command)

            def write_raw(self, command: bytes):
                self.controller.write_raw(command)

            def read(self) -> str:
                return self.controller.read()

            def read_raw(self, memory: int = 100000000) -> bytes:
                return self.controller.read_raw(memory)

        return Driver_VISA

    if connection == 'GPIB':
        class Driver_GPIB(Driver):
            def __init__(self, address: int = 23, board_index: int = 0,
                         **kwargs):
                import Gpib

                self.controller = Gpib.Gpib(int(board_index), int(address))
                Driver.__init__(self)

            def query(self, command: str) -> str:
                self.write(command)
                return self.read()

            def write(self, command: str):
                self.controller.write(command)

            def read(self, length=1000000000) -> str:
                return self.controller.read(length).decode().strip('\n')

            def close(self):
                """WARNING: GPIB closing is automatic at sys.exit() doing it twice results in a gpib error"""
                #Gpib.gpib.close(self.controller.id)
                pass

        return Driver_GPIB

    # if connection == 'USB':
    #     class Driver_USB(Driver):
    #         def __init__(self, **kwargs):
    #             import usb
    #             import usb.core
    #             import usb.util

    #             dev = usb.core.find(idVendor=0x104d, idProduct=0x100a)
    #             dev.reset()
    #             dev.set_configuration()
    #             interface = 0
    #             if dev.is_kernel_driver_active(interface) is True:
    #                 dev.detach_kernel_driver(interface)  # tell the kernel to detach
    #                 usb.util.claim_interface(dev, interface)  # claim the device

    #             cfg = dev.get_active_configuration()
    #             intf = cfg[(0,0)]
    #             self.ep_out = usb.util.find_descriptor(intf, custom_match=lambda e: usb.util.endpoint_direction(e.bEndpointAddress) == usb.util.ENDPOINT_OUT)
    #             self.ep_in = usb.util.find_descriptor(intf, custom_match=lambda e: usb.util.endpoint_direction(e.bEndpointAddress) == usb.util.ENDPOINT_IN)

    #             assert self.ep_out is not None
    #             assert self.ep_in is not None

    #             Driver.__init__(self)

    #         def write(self, query):
    #             self.string = query + '\r\n'
    #             self.ep_out.write(self.string)

    #         def read(self):
    #             rep = self.ep_in.read(64)
    #             const = ''.join(chr(i) for i in rep)
    #             const = const[:const.find('\r\n')]
    #             return const

    #     return Driver_USB

    if connection == 'SOCKET':
        class Driver_SOCKET(Driver):

            BUFFER_SIZE: int = 40000

            def __init__(self, address: str = '192.168.0.8', port: int = 5005,
                         **kwargs):

                import socket

                self.ADDRESS = address
                self.PORT = port

                self.controller = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

                self.s.settimeout(5)

                self.controller.connect((self.ADDRESS, int(self.PORT)))

                Driver.__init__(self, **kwargs)

            def write(self, command: str):
                self.controller.send(command.encode())
                self.controller.recv(self.BUFFER_SIZE)

            def write_raw(self, command: bytes):
                self.controller.send(command)

            def query(self, command: str) -> str:
                self.controller.send(command.encode())
                data = self.controller.recv(self.BUFFER_SIZE)
                return data.decode()

            def read(self, memory: int = BUFFER_SIZE) -> str:
                rep = self.controller.recv(memory).decode()
                return rep

            def read_raw(self, memory: int = BUFFER_SIZE) -> bytes:
                rep = self.controller.recv(memory)
                return rep

            def close(self):
                try: self.controller.close()
                except: pass
                self.controller = None

        return Driver_SOCKET

    if connection == 'TEST':
        class Controller: pass
        class Driver_TEST(Driver):
            def __init__(self, *args, **kwargs):
                try:
                    Driver.__init__(self)
                except:
                    Driver.__init__(self, *args, **kwargs)

                self.controller = Controller()
                self.controller.timeout = 5000

            def write(self, value: str):
                pass
            def write_raw(self, value: bytes):
                pass
            def read(self) -> str:
                return '1'
            def read_raw(self) -> bytes:
                return b'1'
            def query(self, value: str) -> str:
                self.write(value)
                return self.read()

        return Driver_TEST

    return None


def get_module_class(driver_lib: ModuleType, module_name: str) -> Type:
    ''' Returns the class Module_XXX of the provided driver library and module_name'''
    assert module_name in get_module_names(driver_lib)
    return getattr(driver_lib, f'Module_{module_name}')


def explore_driver(instance: Type, _print: bool = True) -> str:
    ''' Displays the list of the methods available in this instance '''
    methods = get_instance_methods(instance)
    s = 'This instance contains the following functions:\n'

    for method in methods:
        s += f' - {method[0]}({",".join(method[1])})\n'

    if _print:
        print(s)
        return None
    return s


def get_instance_methods(instance: Type) -> List[Tuple[str, Type]]:
    ''' Returns the list of all the methods (and their args) in that class '''
    methods = []

    # LEVEL 1
    for name, _ in inspect.getmembers(instance, inspect.ismethod):
        if not name.startswith('_'):
            attr = getattr(instance, name)
            args = list(inspect.signature(attr).parameters)
            methods.append([name, args])

    # LEVEL 2
    for key, val in vars(instance).items():
        if key.startswith('_'): continue
        try:  # explicit to avoid visa and inspect.getmembers issue
            for name, _ in inspect.getmembers(val, inspect.ismethod):
                if not name.startswith('_'):
                    attr = getattr(val, name)
                    args = list(inspect.signature(attr).parameters)
                    methods.append([f'{key}.{name}', args])
        except: pass

    return methods


def get_class_args(clas: Type) -> dict:
    ''' Returns the dictionary of the optional arguments required by a class
    with their default values '''
    signature = inspect.signature(clas)
    return {k: v.default for k, v in signature.parameters.items() if (
        v.default is not inspect.Parameter.empty)}


# =============================================================================
# DRIVERS PATHS
# =============================================================================

def get_driver_path(driver_name: str) -> str:
    ''' Returns the config associated with driver_name '''
    assert isinstance(driver_name, str), "drive_name must be a string."
    assert driver_name in DRIVERS_PATHS, f'Driver {driver_name} not found.'
    return DRIVERS_PATHS[driver_name]['path']


def load_drivers_paths() -> dict:
    ''' Returns a dictionary with:
        - key: name of the driver
        - value: path of the driver python script
    '''
    drivers_paths = {}
    for source_name, source_path in DRIVER_SOURCES.items():
        if not os.path.isdir(source_path):
            print(f"Warning, can't found driver folder: {source_path}")
            continue
        for driver_name in os.listdir(source_path):
            temp_path = os.path.join(source_path, driver_name)
            if (os.path.isdir(temp_path)
                    and f'{driver_name}.py' in os.listdir(temp_path)):
                ## Before, raised error if two identical drivers in different folders, I thought of putting a warning message instead but there was too much printing, now don't say that overwrite path (don't overwrite file).
                # if driver_name in drivers_paths.keys():
                #     print(f"Two drivers where found with the name '{driver_name}', will use path {source_name}")
                # assert driver_name not in drivers_paths.keys(), f"Two drivers where found with the name '{driver_name}'. Each driver must have a unique name."
                drivers_paths[driver_name] = {
                    'path': os.path.join(temp_path, f'{driver_name}.py'),
                    'source': source_name}

    return drivers_paths


def update_drivers_paths():
    ''' Update list of available driver '''
    DRIVERS_PATHS.clear()
    DRIVERS_PATHS.update(load_drivers_paths())



# -*- coding: utf-8 -*-
"""
Created on Thu Apr 13 23:29:33 2017

@author: Quentin Chateiller
quentin.chateiller@c2n.upsaclay.fr

"""

import sys
import queue
import time
import uuid
from typing import Any, Type, Union

import numpy as np
import pandas as pd
from qtpy import QtCore, QtWidgets, QtGui
import pyqtgraph as pg
from pyqtgraph.dockarea.DockArea import DockArea
from pyqtgraph.dockarea.Dock import Dock
from pyqtgraph.console import ConsoleWidget

from .thread import ThreadManager
from .treewidgets import TreeWidgetItemModule
from ..scanning.main import Scanner
from ..GUI_instances import (openPlotter, closePlotter,
                             openAbout, closeAbout,
                             openAddDevice, closeAddDevice,
                             openVariablesMenu, closeVariablesMenu,
                             openPreferences, closePreferences,
                             openDriverInstaller, closeDriverInstaller,
                             closeMonitors,
                             closeSliders,
                             )
from ..icons import icons
from ...paths import PATHS
from ...devices import list_devices, list_loaded_devices, Device, close
from ...elements import Variable as Variable_og
from ...elements import Action
from ...config import get_control_center_config
from ...utilities import boolean, open_file
from ...web import report, doc


class OutputWrapper(QtCore.QObject):
    """ https://stackoverflow.com/questions/19855288/duplicate-stdout-stderr-in-qtextedit-widget """
    outputWritten = QtCore.Signal(object, object)

    def __init__(self, parent, stdout: bool = True,
                 logger_active: bool = False, print_active: bool = True):
        super().__init__(parent)

        if stdout:
            self._stream = sys.stdout
            sys.stdout = self
        else:
            self._stream = sys.stderr
            sys.stderr = self

        self._stdout = stdout
        self.logger_active = logger_active
        self.print_active = print_active

    def write(self, text: str):
        if self.print_active: self._stream.write(text)
        if self.logger_active: self.outputWritten.emit(text, self._stdout)

    def __getattr__(self, name: str) -> Any:
        return getattr(self._stream, name)

    def __del__(self):
        try:
            if self._stdout:
                sys.stdout = self._stream
            else:
                sys.stderr = self._stream
        except AttributeError:
            pass


# Tree widget configuration
class MyQTreeWidget(QtWidgets.QTreeWidget):

    def __init__(self, gui, parent=None):
        self.gui = gui
        super().__init__(parent)

    def startDrag(self, supportedActions):
        dragged_item = self.currentItem()

        if isinstance(dragged_item, TreeWidgetItemModule):
            return None

        if self.gui.scanner is not None:
            self.gui.scanner.setWindowState(
                self.gui.scanner.windowState() & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
            self.gui.scanner.activateWindow()
        super().startDrag(supportedActions)

    def keyPressEvent(self, event):
        if (event.key() == QtCore.Qt.Key_C
                and event.modifiers() == QtCore.Qt.ControlModifier):
            self.copy_item(event)
        else:
            super().keyPressEvent(event)

    def copy_item(self, event):
        if len(self.selectedItems()) == 0:
            super().keyPressEvent(event)
            return None
        item = self.selectedItems()[0]  # assume can select only one item
        if hasattr(item, 'variable'):
            text = item.variable.address()
        elif hasattr(item, 'action'):
            text = item.action.address()
        elif hasattr(item, 'module'):
            if hasattr(item.module, 'address'):
                text = item.module.address()
            else:
                text = item.name
        else:
            print(f'Should not be possible: {item}')
            super().keyPressEvent(event)
            return None

        # Copy the text to the system clipboard
        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setText(text)


class ControlCenter(QtWidgets.QMainWindow):
    """ Main Qt window, Used to control devices, open scanner... """

    def __init__(self):

        # Set up the user interface.
        super().__init__()

        # Window configuration
        self.setWindowTitle("AUTOLAB - Control Panel")
        self.setWindowIcon(icons['autolab'])
        self.setFocus()
        self.activateWindow()
        self.resize(570, 700)

        self.close_device_on_exit = True

        # Thread manager
        self.threadManager = ThreadManager(self)
        self.threadDeviceDict = {}
        self.threadItemDict = {}

        # Scanner
        self.scanner = None

        # Timer for device instantiation
        self.timerDevice = QtCore.QTimer(self)
        self.timerDevice.setInterval(50) # ms
        self.timerDevice.timeout.connect(self.timerAction)

        # queue and timer to add/remove plot from driver
        self.queue_driver = queue.Queue()
        self.dict_widget = {}
        self.timerQueue = QtCore.QTimer(self)
        self.timerQueue.setInterval(int(50)) # ms
        self.timerQueue.timeout.connect(self._queueDriverHandler)
        self._stop_timerQueue = False

        # Import Autolab config
        control_center_config = get_control_center_config()
        logger_active = boolean(control_center_config['logger'])
        console_active = boolean(control_center_config['console'])

        self._logger_area = None
        self._console_area = None
        self.stdout = None
        self.stderr = None

        self.init_ui()

        if logger_active:
            self.add_logger()

        if console_active:
            self.add_console()

    def init_ui(self):
        # Main frame configuration: centralWidget(verticalLayout(splitter(tree)))
        centralWidget = QtWidgets.QWidget()
        self.setCentralWidget(centralWidget)

        verticalLayout = QtWidgets.QVBoxLayout(centralWidget)
        verticalLayout.setSpacing(0)
        verticalLayout.setContentsMargins(0,0,0,0)

        self.splitter = QtWidgets.QSplitter()
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        verticalLayout.addWidget(self.splitter)

        self.tree = MyQTreeWidget(self)
        self.tree.last_drag = None
        self.tree.setHeaderLabels(['Objects', 'Type', 'Actions', 'Values', ''])
        self.tree.header().setDefaultAlignment(QtCore.Qt.AlignCenter)
        self.tree.header().setMinimumSectionSize(15)
        self.tree.header().resizeSection(0, 200)
        self.tree.header().resizeSection(4, 15)
        self.tree.setDragDropMode(QtWidgets.QAbstractItemView.DragOnly)
        self.tree.header().setStretchLastSection(False)
        self.tree.itemClicked.connect(self.itemClicked)
        self.tree.itemPressed.connect(self.itemPressed)
        self.tree.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.tree.customContextMenuRequested.connect(self.rightClick)
        self.tree.setAlternatingRowColors(True)

        self.splitter.addWidget(self.tree)

        closeWidget = QtWidgets.QCheckBox('Disconnect devices on exit')
        closeWidget.setChecked(self.close_device_on_exit)
        closeWidget.stateChanged.connect(
            lambda state: setattr(self, 'close_device_on_exit', state))

        horizontalLayout = QtWidgets.QHBoxLayout()
        horizontalLayout.addStretch()
        horizontalLayout.addWidget(closeWidget)

        verticalLayout.addLayout(horizontalLayout)

        self.menuBar = self.menuBar()
        self.statusBar = self.statusBar()

        # Menu
        guiMenu = self.menuBar.addMenu('Panels')

        scanAction = guiMenu.addAction('Scanner')
        scanAction.setIcon(icons['scanner'])
        scanAction.triggered.connect(self.openScanner)
        scanAction.setStatusTip('Open the scanner in another window')

        plotAction = guiMenu.addAction('Plotter')
        plotAction.setIcon(icons['plotter'])
        plotAction.triggered.connect(lambda: openPlotter(has_parent=True))
        plotAction.setStatusTip('Open the plotter in another window')

        variablesMenuAction = guiMenu.addAction('Variables')
        variablesMenuAction.setIcon(icons['variables'])
        variablesMenuAction.triggered.connect(lambda: openVariablesMenu(True))
        variablesMenuAction.setStatusTip("Open the variable menu in another window")

        settingsMenu = self.menuBar.addMenu('Settings')

        Preferences = settingsMenu.addAction('Preferences')
        Preferences.setIcon(icons['preference'])
        Preferences.triggered.connect(lambda: openPreferences(gui=self))
        Preferences.setStatusTip("Open the preferences window")

        userFolder = settingsMenu.addAction('User folder')
        userFolder.setIcon(icons['folder'])
        userFolder.triggered.connect(self.openUserFolder)
        userFolder.setStatusTip("Open the user folder")

        settingsMenu.addSeparator()

        devicesConfig = settingsMenu.addAction('Devices config')
        devicesConfig.setIcon(icons['file'])
        devicesConfig.triggered.connect(self.openDevicesConfig)
        devicesConfig.setStatusTip("Open the devices configuration file")

        downloadDriverAction = settingsMenu.addAction('Download drivers')
        downloadDriverAction.setIcon(icons['read-save'])
        downloadDriverAction.triggered.connect(lambda: openDriverInstaller(gui=self))
        downloadDriverAction.setStatusTip("Open the utility to download drivers")

        settingsMenu.addSeparator()

        refreshAction = settingsMenu.addAction('Refresh devices')
        refreshAction.setIcon(icons['reload'])
        refreshAction.triggered.connect(self.initialize)
        refreshAction.setStatusTip('Reload devices setting')

        helpMenu = self.menuBar.addMenu('Help')

        reportAction = helpMenu.addAction('Report bugs / suggestions')
        reportAction.setIcon(icons['github'])
        reportAction.triggered.connect(report)
        reportAction.setStatusTip('Open the issue webpage of this project on GitHub')

        helpMenu.addSeparator()

        helpAction = helpMenu.addAction('Documentation (Online)')
        helpAction.setIcon(icons['readthedocs'])
        helpAction.triggered.connect(lambda: doc(True))
        helpAction.setStatusTip('Open the documentation on Read The Docs website')

        helpActionOffline = helpMenu.addAction('Documentation (Offline)')
        helpActionOffline.setIcon(icons['pdf'])
        helpActionOffline.triggered.connect(lambda: doc(False))
        helpActionOffline.setStatusTip(
            'Open the pdf documentation form local file ' \
            '(may no be up-to-day with online version)')

        helpMenu.addSeparator()

        aboutAction = helpMenu.addAction('About Autolab')
        aboutAction.setIcon(icons['autolab'])
        aboutAction.triggered.connect(lambda: openAbout(self))
        aboutAction.setStatusTip('Information about Autolab')
        # / Menu

    def update_handle(self):
        for splitter in (self.splitter, ):
            for i in range(splitter.count()):
                handle = splitter.handle(i)
                handle.setStyleSheet("background-color: #DDDDDD;")
                handle.installEventFilter(self)

    def activate_logger(self, activate: bool):
        if activate:
            self.add_logger()
        else:
            self.remove_logger()

    def add_logger(self):
        self.remove_logger()

        logger_area = DockArea(self)
        self._logger_area = logger_area
        self.splitter.addWidget(logger_area)

        logger_dock = Dock("Logger")
        logger_area.addDock(logger_dock, 'bottom')

        self.logger = QtWidgets.QTextBrowser(self)
        self.loggerDefaultColor = self.logger.textColor()

        # Set logger
        control_center_config = get_control_center_config()
        print_active = boolean(control_center_config['print'])
        self.stdout = OutputWrapper(self, True, True, print_active)
        self.stderr = OutputWrapper(self, False, True, print_active)
        self.stdout.outputWritten.connect(self.handleOutput)
        self.stderr.outputWritten.connect(self.handleOutput)

        logger_dock.addWidget(self.logger)

        self.update_handle()

    def remove_logger(self):
        if self._logger_area:
            self._logger_area.deleteLater()
            self._logger_area = None

            if self.stdout:
                sys.stdout = self.stdout._stream
                self.stdout = None

            if self.stderr:
                sys.stderr = self.stderr._stream
                self.stderr = None

    def activate_console(self, activate: bool):
        if activate:
            self.add_console()
        else:
            self.remove_console()

    def add_console(self):
        self.remove_console()

        import autolab
        namespace = {'np': np, 'pd': pd, 'autolab': autolab}
        text = """ Packages imported: autolab, numpy as np, pandas as pd.\n"""
        console_area = DockArea(self)
        self._console_area = console_area
        self.splitter.addWidget(console_area)

        console_dock = Dock("Console")
        console_area.addDock(console_dock, 'bottom')

        console_widget = ConsoleWidget(namespace=namespace, text=text)
        console_dock.addWidget(console_widget)

        self.update_handle()

    def remove_console(self):
        if self._console_area:
            self._console_area.deleteLater()
            self._console_area = None

    def eventFilter(self, obj, event):
        if event.type() == QtCore.QEvent.Enter:
            obj.setStyleSheet("background-color: #AAAAAA;")  # Hover color
        elif event.type() == QtCore.QEvent.Leave:
            obj.setStyleSheet("background-color: #DDDDDD;")  # Normal color
        return super().eventFilter(obj, event)

    def createWidget(self, widget: Type, *args, **kwargs):
        """ Function used by a driver to add a widget.
        Mainly used to open a figure outside the GUI from a driver. """
        unique_name = str(uuid.uuid4())
        self.queue_driver.put(('create', unique_name, widget, args, kwargs))

        start = time.time()
        while True:
            widget_created = self.dict_widget.get(unique_name)

            if widget_created: return widget_created
            time.sleep(0.01)
            if (time.time() - start) > 1:
                print(f"Warning: Importation of {widget} too long, skip it",
                      file=sys.stderr)
                return None

    def removeWidget(self, widget: Type):
        """ Function used by a driver to remove a widget record from GUI """
        self.queue_driver.put(('remove', "", widget, (), {}))

    def _queueDriverHandler(self):
        """ Adds/Removes plot from a driver requests throught a queue """
        while not self.queue_driver.empty():
            data = self.queue_driver.get()
            action = data[0]
            widget_name = data[1]
            widget = data[2]
            args = data[3]
            kwargs = data[4]

            if action == 'create':
                widget = widget(*args, **kwargs)
                self.dict_widget[widget_name] = widget
            elif action == "remove":
                d = self.dict_widget
                if widget is not None:
                    widget_pos = list(d.values()).index(widget)
                    if widget_pos is not None:
                        widget_name = list(d)[widget_pos]
                        widget = d.get(widget_name)
                        if widget is not None: d.pop(widget_name)

        if self._stop_timerQueue:
            self.timerQueue.stop()
            self._stop_timerQueue = False

    def handleOutput(self, text: str, stdout: bool):
        if not stdout: self.logger.setTextColor(QtCore.Qt.red)
        self.logger.insertPlainText(text)
        self.logger.moveCursor(QtGui.QTextCursor.End)
        self.logger.setTextColor(self.loggerDefaultColor)

    def timerAction(self):
        """ This function checks if a module has been loaded and put to the queue.
        If it has been, associate item and module """
        threadItemDictTemp = self.threadItemDict.copy()
        threadDeviceDictTemp = self.threadDeviceDict.copy()

        for item_id, module in threadDeviceDictTemp.items():
            item = threadItemDictTemp[item_id]

            self.associate(item, module)
            item.setExpanded(True)

            # if hasattr(module.instance, 'gui'):  # Can't just start timer here because driver created before. But could reduce timer interval here if want to start with longer interval while no driver openned
            #     self.timerQueue.setInterval(int(50)) # ms

            self.threadItemDict.pop(item_id)
            self.threadDeviceDict.pop(item_id)

        if len(threadItemDictTemp) == 0:
            self.timerDevice.stop()

    def initialize(self):
        """ This function will create the first items in the tree, but will
        associate only the ones already loaded in autolab """
        self.tree.clear()
        try:
            devices_name = list_devices()
        except Exception as e:
            self.setStatus(f'Error {e}', 10000, False)
        else:
            for dev_name in devices_name:
                item = TreeWidgetItemModule(self.tree, dev_name, self)

                for i in range(5):
                    item.setBackground(i, QtGui.QColor('#9EB7F5'))  # blue

                if dev_name in list_loaded_devices(): self.itemClicked(item)

    def setStatus(self, message: str, timeout: int = 0, stdout: bool = True):
        """ Modify the message displayed in the status bar and add error message to logger """
        self.statusBar.showMessage(message, timeout)
        if not stdout: print(message, file=sys.stderr)

    def clearStatus(self):
        """ Erase the message displayed in the status bar """
        self.setStatus('')

    def rightClick(self, position: QtCore.QPoint):
        """ Function called when a right click has been detected in the tree """
        item = self.tree.itemAt(position)
        if hasattr(item, 'menu'): item.menu(position)
        elif item is None: self.addDeviceMenu(position)

    def addDeviceMenu(self, position: QtCore.QPoint):
        """ Open menu to ask if want to add new device """
        menu = QtWidgets.QMenu()
        addDeviceChoice = menu.addAction('Add Device')
        addDeviceChoice.setIcon(icons['add'])

        choice = menu.exec_(self.tree.viewport().mapToGlobal(position))

        if choice == addDeviceChoice:
            openAddDevice(gui=self)

    def itemClicked(self, item: QtWidgets.QTreeWidgetItem):
        """ Function called when a normal click has been detected in the tree.
            Check the association if it is a main item """
        if (item.parent() is None and not item.loaded
                and id(item) not in self.threadItemDict):
            self.threadItemDict[id(item)] = item  # needed before start of timer to avoid bad timing and to stop thread before loading is done
            item.setValueKnownState(0.5)  # yellow color
            self.threadManager.start(item, 'load')  # load device and add it to queue for timer to associate it later (doesn't block gui while device is openning)
            self.timerDevice.start()
            self.timerQueue.start()

    def itemCanceled(self, item):
        """ Cancel the device openning. Can be used to avoid GUI blocking
        for devices with infinite loading issue """
        if id(item) in self.threadManager.threads_conn:
            tid = self.threadManager.threads_conn[id(item)]
            self.threadManager.threads[tid].endSignal.emit(
                f'Cancel loading device {item.name}')
            self.threadManager.threads[tid].terminate()

    def itemPressed(self, item: QtWidgets.QTreeWidgetItem):
        """ Function called when a click (not released) has been detected in the tree.
            Store last dragged variable in tree so scanner can know it when it is dropped there.
            """
        if hasattr(item, "module"):
            if item.is_not_submodule: self.tree.last_drag = item.name
            else: self.tree.last_drag = None
        elif hasattr(item, "variable"): self.tree.last_drag = item.variable
        elif hasattr(item, "action"): self.tree.last_drag = item.action
        else: self.tree.last_drag = None

    def associate(self, item: QtWidgets.QTreeWidgetItem, module: Device):
        """ Function called to associate a main module to one item in the tree """
        # load the entire module (submodules, variables, actions)
        item.load(module)

        read_init_list = module._read_init_list
        for mod in module._mod.values():
            read_init_list.extend(mod._read_init_list)
        for variable in read_init_list:
            try:
                variable()
            except:
                self.setStatus(f"Can't read variable {variable.address()} on instantiation",
                               10000, False)

    def openScanner(self):
        """ This function open the scanner. """
        # If the scanner is not already running, create one
        if self.scanner is None:
            self.scanner = Scanner(self)
            self.scanner.show()
            self.scanner.activateWindow()
            self.activateWindow() # Put main window back to the front
        # If the scanner is already running, just make as the front window
        else:
            self.scanner.setWindowState(
                self.scanner.windowState() & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
            self.scanner.activateWindow()

    def clearScanner(self):
        """ This clear the scanner instance reference when quitted """
        self.scanner = None

    @staticmethod
    def openDevicesConfig():
        """ Open the devices configuration file """
        open_file(PATHS['devices_config'])

    @staticmethod
    def openUserFolder():
        """ Open the user folder """
        open_file(PATHS['user_folder'])

    def setScanParameter(self, recipe_name: str, param_name: str,
                         variable: Variable_og):
        """ Set the selected variable has parameter for the recipe """
        if self.scanner is None:
            self.openScanner()

        self.scanner.configManager.setParameter(recipe_name, param_name, variable)

    def addStepToScanRecipe(self, recipe_name: str, stepType: str,
                            element: Union[Variable_og, Action],
                            value: Any = None):
        """ Add the selected variable has a step for the recipe """
        if self.scanner is None:
            self.openScanner()

        self.scanner.configManager.addRecipeStep(
            recipe_name, stepType, element, value=value)

    def getRecipeName(self) -> str:
        """ Returns the name of the recipe that will receive the variables
        from the control center """
        if self.scanner is None:
            self.openScanner()

        return self.scanner.selectRecipe_comboBox.currentText()

    def getParameterName(self) -> str:
        """ Returns the name of the parameter that will receive the variables
        from the control center """
        if self.scanner is None:
            self.openScanner()

        return self.scanner.selectParameter_comboBox.currentText()

    def closeEvent(self, event):
        """ This function does some steps before the window is really killed """
        if self.scanner:
            self.scanner.close()

            if self.scanner is not None and self.scanner.isVisible():
                event.ignore()
                return None

        closePlotter()
        closeAbout()
        closeAddDevice()
        closeMonitors()
        closeSliders()
        closeVariablesMenu()
        closePreferences()
        closeDriverInstaller()

        if self.close_device_on_exit:
            close()  # close all devices

        self.remove_logger()
        self.remove_console()

        self.timerDevice.stop()
        self.timerQueue.stop()

        try:
            # Prevent 'RuntimeError: wrapped C/C++ object of type ViewBox has been deleted' when reloading gui
            for view in pg.ViewBox.AllViews.copy().keys():
                pg.ViewBox.forgetView(id(view), view)

            pg.ViewBox.quit()
        except: pass

        for children in self.findChildren(QtWidgets.QWidget):
            children.deleteLater()

        super().closeEvent(event)

        QtWidgets.QApplication.quit()  # close the app

        # OPTIMIZE: don't know if should erase variables on exit or not.
        # Currently decided to only erase variables defined by scanner itself,
        # keeping the ones defined by user.
        # VARIABLES.clear()  # reset variables defined in the GUI



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 29 18:29:07 2019

@author: qchat
"""

from typing import Any, Union
import os

import pandas as pd
import numpy as np
from qtpy import QtCore, QtWidgets, QtGui

from ..icons import icons
from ..GUI_utilities import (MyLineEdit, MyInputDialog, MyQCheckBox, MyQComboBox,
                             RecipeMenu, qt_object_exists)
from ..GUI_instances import openMonitor, openSlider, openPlotter, openAddDevice
from ...paths import PATHS
from ...config import get_control_center_config
from ...variables import eval_variable, has_eval
from ...devices import close, list_loaded_devices
from ...utilities import (SUPPORTED_EXTENSION, str_to_array, array_to_str,
                          dataframe_to_str, str_to_dataframe, create_array,
                          str_to_tuple)


class TreeWidgetItemModule(QtWidgets.QTreeWidgetItem):
    """ This class represents a module in an item of the tree """

    def __init__(self, itemParent, name, gui):

        self.name = name
        self.module = None
        self.loaded = False
        self.gui = gui
        self.is_not_submodule = isinstance(gui.tree, type(itemParent))

        if self.is_not_submodule:
            super().__init__(itemParent, [name, 'Device'])
            self.indicator = QtWidgets.QLabel()
            self.indicator.setToolTip('Device not instantiated yet')
            self.gui.tree.setItemWidget(self, 4, self.indicator)
        else:
            super().__init__(itemParent, [name, 'Module'])

        self.setTextAlignment(1, QtCore.Qt.AlignHCenter)

    def load(self, module):
        """ This function loads the entire module (submodules, variables, actions) """
        self.module = module

        # Submodules
        subModuleNames = self.module.list_modules()
        for subModuleName in subModuleNames:
            subModule = getattr(self.module, subModuleName)
            item = TreeWidgetItemModule(self, subModuleName, self.gui)
            item.load(subModule)

        # Variables
        varNames = self.module.list_variables()
        for varName in varNames:
            variable = getattr(self.module,varName)
            TreeWidgetItemVariable(self, variable, self.gui)

        # Actions
        actNames = self.module.list_actions()
        for actName in actNames:
            action = getattr(self.module, actName)
            TreeWidgetItemAction(self, action, self.gui)

        # Change loaded status
        self.loaded = True

        # Tooltip
        if self.module._help is not None: self.setToolTip(0, self.module._help)

    def menu(self, position: QtCore.QPoint):
        """ This function provides the menu when the user right click on an item """
        if self.is_not_submodule:
            if self.loaded:
                menu = QtWidgets.QMenu()
                disconnectDevice = menu.addAction(f"Disconnect {self.name}")
                disconnectDevice.setIcon(icons['disconnect'])

                choice = menu.exec_(self.gui.tree.viewport().mapToGlobal(position))

                if choice == disconnectDevice:
                    close(self.name)

                    for i in range(self.childCount()):
                        self.removeChild(self.child(0))

                    self.loaded = False
                    self.setValueKnownState(False)

                if not list_loaded_devices():
                    self.gui._stop_timerQueue = True

            elif id(self) in self.gui.threadManager.threads_conn:
                menu = QtWidgets.QMenu()
                cancelDevice = menu.addAction('Cancel loading')
                cancelDevice.setIcon(icons['disconnect'])

                choice = menu.exec_(self.gui.tree.viewport().mapToGlobal(position))

                if choice == cancelDevice:
                    self.gui.itemCanceled(self)
            else:
                menu = QtWidgets.QMenu()
                modifyDeviceChoice = menu.addAction('Modify device')
                modifyDeviceChoice.setIcon(icons['rename'])

                choice = menu.exec_(self.gui.tree.viewport().mapToGlobal(position))

                if choice == modifyDeviceChoice:
                    openAddDevice(gui=self.gui, name=self.name)

    def setValueKnownState(self, state: Union[bool, float]):
        """ Turn the color of the indicator depending of the known state of the value """
        if state == 0.5:
            self.indicator.setStyleSheet("background-color:#FFFF00")  # yellow
            self.indicator.setToolTip('Device is being instantiated')
        elif state == -1:
            self.indicator.setStyleSheet("background-color:#FF0000")  # red
            self.indicator.setToolTip('Device connection error')
        elif state:
            self.indicator.setStyleSheet("background-color:#70db70")  # green
            self.indicator.setToolTip('Device instantiated')
        else:
            self.indicator.setStyleSheet("background-color:none")  # none
            self.indicator.setToolTip('Device closed')


class TreeWidgetItemAction(QtWidgets.QTreeWidgetItem):
    """ This class represents an action in an item of the tree """

    def __init__(self, itemParent, action, gui):

        displayName = f'{action.name}'
        if action.unit is not None:
            displayName += f' ({action.unit})'

        super().__init__(itemParent, [displayName, 'Action'])
        self.setTextAlignment(1, QtCore.Qt.AlignHCenter)

        self.gui = gui
        self.action = action

        # Import Autolab config
        control_center_config = get_control_center_config()
        self.precision = int(float(control_center_config['precision']))

        if self.action.has_parameter:
            if self.action.type in [int, float, bool, str, bytes, tuple, np.ndarray, pd.DataFrame]:
                self.executable = True
                self.has_value = True
            else:
                self.executable = False
                self.has_value = False
        else:
            self.executable = True
            self.has_value = False

        # Main - Column 2 : actionread button
        if self.executable:
            self.execButton = QtWidgets.QPushButton()
            self.execButton.setText("Execute")
            self.execButton.clicked.connect(self.execute)
            self.gui.tree.setItemWidget(self, 2, self.execButton)

        # Main - Column 3 : QlineEdit if the action has a parameter
        if self.has_value:
            if self.action.type in [int, float, str, bytes, np.ndarray, pd.DataFrame]:
                self.valueWidget = MyLineEdit()
                self.valueWidget.setAlignment(QtCore.Qt.AlignCenter)
                self.gui.tree.setItemWidget(self, 3, self.valueWidget)
                self.valueWidget.returnPressed.connect(self.execute)
                self.valueWidget.textEdited.connect(self.valueEdited)

            ## QCheckbox for boolean variables
            elif self.action.type in [bool]:
                self.valueWidget = MyQCheckBox(self)
                hbox = QtWidgets.QHBoxLayout()
                hbox.addWidget(self.valueWidget)
                hbox.setAlignment(QtCore.Qt.AlignCenter)
                hbox.setSpacing(0)
                hbox.setContentsMargins(0,0,0,0)
                widget = QtWidgets.QFrame()
                widget.setLayout(hbox)

                self.gui.tree.setItemWidget(self, 3, widget)

            ## Combobox for tuples: Tuple[List[str], int]
            elif self.action.type in [tuple]:
                self.valueWidget = MyQComboBox()
                self.valueWidget.wheel = False  # prevent changing value by mistake
                self.valueWidget.key = False
                self.valueWidget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
                self.valueWidget.customContextMenuRequested.connect(self.openInputDialog)

                self.gui.tree.setItemWidget(self, 3, self.valueWidget)

        # Main - column 4 : indicator (status of the actual value : known or not known)
        self.indicator = QtWidgets.QLabel()
        self.gui.tree.setItemWidget(self, 4, self.indicator)

        # Tooltip
        if self.action._help is None: tooltip = 'No help available for this action'
        else: tooltip = self.action._help
        if hasattr(self.action, "type") and self.action.type is not None:
            action_type = str(self.action.type).split("'")[1]
            tooltip += f" ({action_type})"
        self.setToolTip(0, tooltip)

        self.writeSignal = WriteSignal()
        self.writeSignal.writed.connect(self.valueWrited)
        self.action._write_signal = self.writeSignal

    def openInputDialog(self, position: QtCore.QPoint):
        """ Only used for tuple """
        menu = QtWidgets.QMenu()
        modifyTuple = menu.addAction("Modify tuple")
        modifyTuple.setIcon(icons['tuple'])

        choice = menu.exec_(self.valueWidget.mapToGlobal(position))

        if choice == modifyTuple:
            main_dialog = MyInputDialog(self.gui, self.action.address())
            main_dialog.setWindowModality(QtCore.Qt.ApplicationModal)
            if self.action.type in [tuple]:
                main_dialog.setTextValue(str(self.action.value))
            main_dialog.show()

            if main_dialog.exec_() == QtWidgets.QInputDialog.Accepted:
                response = main_dialog.textValue()
            else:
                response = ''

            if qt_object_exists(main_dialog): main_dialog.deleteLater()

            if response != '':
                try:
                    if has_eval(response):
                        response = eval_variable(response)
                    if self.action.type in [tuple]:
                        response = str_to_tuple(str(response))
                except Exception as e:
                    self.gui.setStatus(
                        f"Variable {self.action.address()}: {e}", 10000, False)
                    return None

                self.action.value = response
                self.valueWrited(response)
                self.valueEdited()

    def writeGui(self, value):
        """ This function displays a new value in the GUI """
        if hasattr(self, 'valueWidget') and qt_object_exists(self.valueWidget):  # avoid crash if device closed and try to write gui (if close device before reading finished)
            # Update value
            if self.action.type in [int, float]:
                self.valueWidget.setText(f'{value:.{self.precision}g}') # default is .6g
            elif self.action.type in [str]:
                self.valueWidget.setText(value)
            elif self.action.type in [bytes]:
                self.valueWidget.setText(value.decode())
            elif self.action.type in [bool]:
                self.valueWidget.setChecked(value)
            elif self.action.type in [tuple]:
                items = [self.valueWidget.itemText(i)
                         for i in range(self.valueWidget.count())]
                if value[0] != items:
                    self.valueWidget.clear()
                    self.valueWidget.addItems(value[0])
                self.valueWidget.setCurrentIndex(value[1])
            elif self.action.type in [np.ndarray]:
                self.valueWidget.setText(array_to_str(value))
            elif self.action.type in [pd.DataFrame]:
                self.valueWidget.setText(dataframe_to_str(value))
            else:
                self.valueWidget.setText(value)

    def readGui(self) -> Any:
        """ This function returns the value in good format of the value in the GUI """
        if self.action.type in [int, float, str, bytes, np.ndarray, pd.DataFrame]:
            value = self.valueWidget.text()

            if value == '':
                if self.action.unit in ('open-file', 'save-file', 'filename'):
                    if self.action.unit == "filename":  # TODO: LEGACY (to remove later)
                        self.gui.setStatus("Using 'filename' as unit is depreciated in favor of 'open-file' and 'save-file'" \
                                           f"\nUpdate driver '{self.action.address().split('.')[0]}' to remove this warning",
                                           10000, False)
                        self.action.unit = "open-file"

                    if self.action.unit == "open-file":
                        filename, _ = QtWidgets.QFileDialog.getOpenFileName(
                            self.gui, caption=f"Open file - {self.action.address()}",
                            directory=PATHS['last_folder'],
                            filter=SUPPORTED_EXTENSION)
                    elif self.action.unit == "save-file":
                        filename, _ = QtWidgets.QFileDialog.getSaveFileName(
                            self.gui, caption=f"Save file - {self.action.address()}",
                            directory=PATHS['last_folder'],
                            filter=SUPPORTED_EXTENSION)

                    if filename != '':
                        path = os.path.dirname(filename)
                        PATHS['last_folder'] = path
                        return filename
                    else:
                        self.gui.setStatus(
                            f"Action {self.action.address()} cancel filename selection",
                            10000)
                elif self.action.unit == "user-input":
                    main_dialog = MyInputDialog(self.gui, self.action.address())
                    main_dialog.setWindowModality(QtCore.Qt.ApplicationModal)
                    main_dialog.show()

                    if main_dialog.exec_() == QtWidgets.QInputDialog.Accepted:
                        response = main_dialog.textValue()
                    else:
                        response = ''

                    if qt_object_exists(main_dialog): main_dialog.deleteLater()

                    if response != '':
                        return response
                    else:
                        self.gui.setStatus(
                            f"Action {self.action.address()} cancel user input",
                            10000)
                else:
                    self.gui.setStatus(
                        f"Action {self.action.address()} requires a value for its parameter",
                        10000, False)
            else:
                try:
                    value = eval_variable(value)
                    if self.action.type in [int]:
                        value = int(float(value))
                    if self.action.type in [bytes]:
                        value = value.encode()
                    elif self.action.type in [np.ndarray]:
                        value = str_to_array(value) if isinstance(
                            value, str) else create_array(value)
                    elif self.action.type in [pd.DataFrame]:
                        if isinstance(value, str):
                            value = str_to_dataframe(value)
                    else:
                        value = self.action.type(value)
                    return value
                except Exception as e:
                    self.gui.setStatus(
                        f"Action {self.action.address()}: {e}",
                        10000, False)
        elif self.action.type in [bool]:
            value = self.valueWidget.isChecked()
            return value
        elif self.action.type in [tuple]:
            items = [self.valueWidget.itemText(i)
                     for i in range(self.valueWidget.count())]
            value = (items, self.valueWidget.currentIndex())
            return value

    def setValueKnownState(self, state: Union[bool, float]):
        """ Turn the color of the indicator depending of the known state of the value """
        if state == 0.5:
            self.indicator.setStyleSheet("background-color:#FFFF00")  # yellow
            self.indicator.setToolTip('Value written but not read')
        elif state:
            self.indicator.setStyleSheet("background-color:#70db70")  # green
            self.indicator.setToolTip('Value read')
        else:
            self.indicator.setStyleSheet("background-color:#ff8c1a")  # orange
            self.indicator.setToolTip('Value not up-to-date')

    def execute(self):
        """ Start a new thread to execute the associated action """
        if not self.isDisabled():
            if self.has_value:
                value = self.readGui()
                if value is not None:
                    self.gui.threadManager.start(self, 'execute', value=value)
            else:
                self.gui.threadManager.start(self, 'execute')

    def menu(self, position: QtCore.QPoint):
        """ This function provides the menu when the user right click on an item """
        if not self.isDisabled():
            menu = RecipeMenu(self.gui.scanner)

            scanRecipe = menu.addAnyAction('Do in scan recipe', 'action')

            choice = menu.exec_(self.gui.tree.viewport().mapToGlobal(position))
            if choice is None: choice = menu.selected_action

            if choice == scanRecipe:
                recipe_name = self.gui.getRecipeName()
                value = self.action.value if self.action.type in [tuple] else None
                self.gui.addStepToScanRecipe(
                    recipe_name, 'action', self.action, value=value)

    def valueEdited(self):
        """ Change indicator state when editing action parameter """
        self.setValueKnownState(False)

    def valueWrited(self, value: Any):
        """ Called when action parameter written """
        try:
            if self.has_value:
                self.writeGui(value)
            self.setValueKnownState(0.5)
        except Exception as e:
            self.gui.setStatus(f"SHOULD NOT RAISE ERROR: {e}", 10000, False)


class TreeWidgetItemVariable(QtWidgets.QTreeWidgetItem):
    """ This class represents a variable in an item of the tree """

    def __init__(self, itemParent, variable, gui):

        displayName = f'{variable.name}'
        if variable.unit is not None:
            displayName += f' ({variable.unit})'

        super().__init__(itemParent, [displayName, 'Variable'])
        self.setTextAlignment(1, QtCore.Qt.AlignHCenter)

        self.gui = gui
        self.variable = variable

        # Import Autolab config
        control_center_config = get_control_center_config()
        self.precision = int(float(control_center_config['precision']))

        # Signal creation and associations in autolab devices instances
        self.readSignal = ReadSignal()
        self.readSignal.read.connect(self.writeGui)
        self.variable._read_signal = self.readSignal
        self.writeSignal = WriteSignal()
        self.writeSignal.writed.connect(self.valueWrited)
        self.variable._write_signal = self.writeSignal

        # Main - Column 2 : Creation of a READ button if the variable is readable
        if self.variable.readable and self.variable.type in [
                int, float, bool, str, bytes, tuple, np.ndarray, pd.DataFrame]:
            self.readButton = QtWidgets.QPushButton()
            self.readButton.setText("Read")
            self.readButton.clicked.connect(self.read)

            if not self.variable.writable and self.variable.type in [
                    np.ndarray, pd.DataFrame]:
                self.readButtonCheck = QtWidgets.QCheckBox()
                self.readButtonCheck.stateChanged.connect(
                    self.readButtonCheckEdited)
                self.readButtonCheck.setToolTip(
                    'Toggle reading in text, ' \
                    'careful can truncate data and impact performance')
                self.readButtonCheck.setMaximumWidth(15)

                frameReadButton = QtWidgets.QFrame()
                hbox = QtWidgets.QHBoxLayout(frameReadButton)
                hbox.setSpacing(0)
                hbox.setContentsMargins(0,0,0,0)
                hbox.addWidget(self.readButtonCheck)
                hbox.addWidget(self.readButton)
                self.gui.tree.setItemWidget(self, 2, frameReadButton)
            else:
                self.gui.tree.setItemWidget(self, 2, self.readButton)

        # Main - column 3 : Creation of a VALUE widget, depending on the type

        ## QLineEdit or QLabel
        if self.variable.type in [int, float, str, bytes, np.ndarray, pd.DataFrame]:
            if self.variable.writable:
                self.valueWidget = MyLineEdit()
                self.valueWidget.setAlignment(QtCore.Qt.AlignCenter)
                self.valueWidget.returnPressed.connect(self.write)
                self.valueWidget.textEdited.connect(self.valueEdited)
                self.valueWidget.setMaxLength(10000000)  # default is 32767, not enought for array and dataframe
                # self.valueWidget.setPlaceholderText(self.variable._help)  # Could be nice but take too much place. Maybe add it as option
            elif self.variable.readable:
                self.valueWidget = QtWidgets.QLineEdit()
                self.valueWidget.setMaxLength(10000000)
                self.valueWidget.setReadOnly(True)
                palette = self.valueWidget.palette()
                palette.setColor(QtGui.QPalette.Base,
                                 palette.color(QtGui.QPalette.Base).darker(107))
                self.valueWidget.setPalette(palette)
                self.valueWidget.setAlignment(QtCore.Qt.AlignCenter)
            else:
                self.valueWidget = QtWidgets.QLabel()
                self.valueWidget.setAlignment(QtCore.Qt.AlignCenter)

            self.gui.tree.setItemWidget(self, 3, self.valueWidget)

        ## QCheckbox for boolean variables
        elif self.variable.type in [bool]:
            self.valueWidget = MyQCheckBox(self)
            hbox = QtWidgets.QHBoxLayout()
            hbox.addWidget(self.valueWidget)
            hbox.setAlignment(QtCore.Qt.AlignCenter)
            hbox.setSpacing(0)
            hbox.setContentsMargins(0,0,0,0)
            widget = QtWidgets.QFrame()
            widget.setLayout(hbox)
            if not self.variable.writable:  # Disable interaction is not writable
                self.valueWidget.setEnabled(False)

            self.gui.tree.setItemWidget(self, 3, widget)

        ## Combobox for tuples: Tuple[List[str], int]
        elif self.variable.type in [tuple]:
            if self.variable.writable:
                self.valueWidget = MyQComboBox()
                self.valueWidget.wheel = False  # prevent changing value by mistake
                self.valueWidget.key = False
                self.valueWidget.activated.connect(self.write)
                self.valueWidget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
                self.valueWidget.customContextMenuRequested.connect(self.openInputDialog)
            elif self.variable.readable:
                self.valueWidget = MyQComboBox()
                self.valueWidget.readonly = True
            else:
                self.valueWidget = QtWidgets.QLabel()
                self.valueWidget.setAlignment(QtCore.Qt.AlignCenter)

            self.gui.tree.setItemWidget(self, 3, self.valueWidget)

        # Main - column 4 : indicator (status of the actual value : known or not known)
        self.indicator = QtWidgets.QLabel()
        self.gui.tree.setItemWidget(self, 4, self.indicator)

        # Tooltip
        if self.variable._help is None: tooltip = 'No help available for this variable'
        else: tooltip = self.variable._help
        if hasattr(self.variable, "type"):
            variable_type = str(self.variable.type).split("'")[1]
            tooltip += f" ({variable_type})"
        self.setToolTip(0, tooltip)

        # disable read button if array/dataframe
        if hasattr(self, 'readButtonCheck'): self.readButtonCheckEdited()

    def openInputDialog(self, position: QtCore.QPoint):
        """ Only used for tuple """
        menu = QtWidgets.QMenu()
        modifyTuple = menu.addAction("Modify tuple")
        modifyTuple.setIcon(icons['tuple'])

        choice = menu.exec_(self.valueWidget.mapToGlobal(position))

        if choice == modifyTuple:
            main_dialog = MyInputDialog(self.gui, self.variable.address())
            main_dialog.setWindowModality(QtCore.Qt.ApplicationModal)
            if self.variable.type in [tuple]:
                main_dialog.setTextValue(str(self.variable.value))
            main_dialog.show()

            if main_dialog.exec_() == QtWidgets.QInputDialog.Accepted:
                response = main_dialog.textValue()
            else:
                response = ''

            if qt_object_exists(main_dialog): main_dialog.deleteLater()

            if response != '':
                try:
                    if has_eval(response):
                        response = eval_variable(response)
                    if self.variable.type in [tuple]:
                        response = str_to_tuple(str(response))
                except Exception as e:
                    self.gui.setStatus(
                        f"Variable {self.variable.address()}: {e}", 10000, False)
                    return None

                self.variable(response)

                if self.variable.readable:
                    self.variable()

    def writeGui(self, value):
        """ This function displays a new value in the GUI """
        if hasattr(self, 'valueWidget') and qt_object_exists(self.valueWidget):  # avoid crash if device closed and try to write gui (if close device before reading finished)
            # Update value
            if self.variable.numerical:
                self.valueWidget.setText(f'{value:.{self.precision}g}') # default is .6g
            elif self.variable.type in [str]:
                self.valueWidget.setText(value)
            elif self.variable.type in [bytes]:
                self.valueWidget.setText(value.decode())
            elif self.variable.type in [bool]:
                self.valueWidget.setChecked(value)
            elif self.variable.type in [tuple]:
                items = [self.valueWidget.itemText(i)
                         for i in range(self.valueWidget.count())]
                if value[0] != items:
                    self.valueWidget.clear()
                    self.valueWidget.addItems(value[0])
                self.valueWidget.setCurrentIndex(value[1])
            elif self.variable.type in [np.ndarray, pd.DataFrame]:
                if self.variable.writable or self.readButtonCheck.isChecked():
                    if self.variable.type in [np.ndarray]:
                        self.valueWidget.setText(array_to_str(value))
                    if self.variable.type in [pd.DataFrame]:
                        self.valueWidget.setText(dataframe_to_str(value))
                # else:
                #     self.valueWidget.setText('')

            # Change indicator light to green
            if self.variable.type in [
                    int, float, bool, str, bytes, tuple, np.ndarray, pd.DataFrame]:
                self.setValueKnownState(True)

    def readGui(self):
        """ This function returns the value in good format of the value in the GUI """
        if self.variable.type in [int, float, str, bytes, np.ndarray, pd.DataFrame]:
            value = self.valueWidget.text()
            if value == '':
                self.gui.setStatus(
                    f"Variable {self.variable.address()} requires a value to be set",
                    10000, False)
            else:
                try:
                    value = eval_variable(value)
                    if self.variable.type in [int]:
                        value = int(float(value))
                    if self.variable.type in [bytes]:
                        value = value.encode()
                    elif self.variable.type in [np.ndarray]:
                        if isinstance(value, str): value = str_to_array(value)
                        else: value = create_array(value)
                    elif self.variable.type in [pd.DataFrame]:
                        if isinstance(value, str): value = str_to_dataframe(value)
                    else:
                        value = self.variable.type(value)
                    return value
                except Exception as e:
                    self.gui.setStatus(
                        f"Variable {self.variable.address()}: {e}",
                        10000, False)

        elif self.variable.type in [bool]:
            value = self.valueWidget.isChecked()
            return value
        elif self.variable.type in [tuple]:
            items = [self.valueWidget.itemText(i)
                     for i in range(self.valueWidget.count())]
            value = (items, self.valueWidget.currentIndex())
            return value

    def setValueKnownState(self, state: Union[bool, float]):
        """ Turn the color of the indicator depending of the known state of the value """
        if state == 0.5:
            self.indicator.setStyleSheet("background-color:#FFFF00")  # yellow
            self.indicator.setToolTip('Value written but not read')
        elif state:
            self.indicator.setStyleSheet("background-color:#70db70")  # green
            self.indicator.setToolTip('Value read')
        else:
            self.indicator.setStyleSheet("background-color:#ff8c1a")  # orange
            self.indicator.setToolTip('Value not up-to-date')

    def read(self):
        """ Start a new thread to READ the associated variable """
        if not self.isDisabled():
            self.setValueKnownState(False)
            self.gui.threadManager.start(self, 'read')

    def write(self):
        """ Start a new thread to WRITE the associated variable """
        if not self.isDisabled():
            value = self.readGui()
            if value is not None:
                self.gui.threadManager.start(self, 'write', value=value)

    def valueWrited(self, value: Any):
        """ Function call when the value displayed in not sure anymore.
            The value has been modified either in the GUI (but not sent)
            or by command line.
            If variable not readable, write the value sent to the GUI """
        # BUG: I got an error when changing emit_write to set value, need to reproduce it
        try:
            self.writeGui(value)
            self.setValueKnownState(0.5)
        except Exception as e:
            self.gui.setStatus(f"SHOULD NOT RAISE ERROR: {e}", 10000, False)

    def valueEdited(self):
        """ Function call when the value displayed in not sure anymore.
            The value has been modified either in the GUI (but not sent)
            or by command line """
        self.setValueKnownState(False)

    def readButtonCheckEdited(self):
        state = bool(self.readButtonCheck.isChecked())
        self.readButton.setEnabled(state)

    def menu(self, position: QtCore.QPoint):
        """ This function provides the menu when the user right click on an item """
        if not self.isDisabled():
            menu = RecipeMenu(self.gui.scanner)
            monitoringAction = menu.addAction("Start monitoring")
            monitoringAction.setIcon(icons['monitor'])
            plottingAction = menu.addAction("Capture to plotter")
            plottingAction.setIcon(icons['plotter'])
            menu.addSeparator()
            sliderAction = menu.addAction("Create a slider")
            sliderAction.setIcon(icons['slider'])
            menu.addSeparator()

            scanParameterAction = menu.addAnyAction(
                'Set as scan parameter', 'parameter', param_menu_active=True)
            scanMeasureStepAction = menu.addAnyAction(
                'Measure in scan recipe', 'measure')
            scanSetStepAction = menu.addAnyAction(
                'Set value in scan recipe', 'write')

            menu.addSeparator()
            saveAction = menu.addAction("Read and save as...")
            saveAction.setIcon(icons['read-save'])

            monitoringAction.setEnabled(
                self.variable.readable and self.variable.type in [
                    int, float, np.ndarray, pd.DataFrame])
            plottingAction.setEnabled(monitoringAction.isEnabled())
            sliderAction.setEnabled((self.variable.writable
                                     and self.variable.type in [int, float]))
            scanParameterAction.setEnabled(self.variable.parameter_allowed)
            scanMeasureStepAction.setEnabled(self.variable.readable)
            saveAction.setEnabled(self.variable.readable)
            scanSetStepAction.setEnabled(self.variable.writable)

            choice = menu.exec_(self.gui.tree.viewport().mapToGlobal(position))
            if choice is None: choice = menu.selected_action

            if choice == monitoringAction:
                openMonitor(self.variable, has_parent=True)
            if choice == plottingAction:
                openPlotter(variable=self.variable, has_parent=True)
            elif choice == sliderAction:
                openSlider(self.variable, gui=self.gui, item=self)
            elif choice == scanParameterAction:
                recipe_name = self.gui.getRecipeName()
                param_name = self.gui.getParameterName()
                self.gui.setScanParameter(recipe_name, param_name, self.variable)
            elif choice == scanMeasureStepAction:
                recipe_name = self.gui.getRecipeName()
                self.gui.addStepToScanRecipe(recipe_name, 'measure', self.variable)
            elif choice == scanSetStepAction:
                recipe_name = self.gui.getRecipeName()
                value = self.variable.value if self.variable.type in [tuple] else None
                self.gui.addStepToScanRecipe(
                    recipe_name, 'set', self.variable, value=value)
            elif choice == saveAction: self.saveValue()

    def saveValue(self):
        """ Prompt user for filename to save data of the variable """
        filename, _ = QtWidgets.QFileDialog.getSaveFileName(
            self.gui, f"Save {self.variable.address()} value", os.path.join(
                PATHS['last_folder'], f'{self.variable.address()}.txt'),
            filter=SUPPORTED_EXTENSION)

        path = os.path.dirname(filename)
        if path != '':
            PATHS['last_folder'] = path
            try:
                self.gui.setStatus(
                    f"Saving value of {self.variable.address()}...", 5000)
                self.variable.save(filename)
                self.gui.setStatus(
                    f"Value of {self.variable.address()} successfully read " \
                    f"and save at {filename}", 5000)
            except Exception as e:
                self.gui.setStatus(f"An error occured: {e}", 10000, False)


# Signals can be emitted only from QObjects
# These class provides convenient ways to use signals
class ReadSignal(QtCore.QObject):
    read = QtCore.Signal(object)
    def emit_read(self, value):
        self.read.emit(value)

class WriteSignal(QtCore.QObject):
    writed = QtCore.Signal(object)
    def emit_write(self, value):
        self.writed.emit(value)



# -*- coding: utf-8 -*-
"""
Created on Mon Jan 15 14:42:16 2024

@author: Jonathan
"""

from typing import List, Union

import numpy as np
import pandas as pd
from qtpy import QtCore, QtWidgets, QtGui

from ..icons import icons
from ...utilities import clean_string, array_to_str, dataframe_to_str
from ...elements import Variable as Variable_og
from ...elements import Action
from ...variables import has_eval
from ...config import get_scanner_config


class MyQTreeWidget(QtWidgets.QTreeWidget):
    customMimeType = "autolab/MyQTreeWidget-selectedItems"

    reorderSignal = QtCore.Signal(object)

    def __init__(self, parent: QtWidgets.QFrame,
                 gui: QtWidgets.QMainWindow, recipe_name: str):

        self.recipe_name = recipe_name
        self.scanner = gui  # gui is scanner
        super().__init__(parent)
        self.setAcceptDrops(True)

    def mimeTypes(self) -> QtWidgets.QTreeWidget.mimeTypes:
        """Based on https://gist.github.com/eyllanesc/42bcda52a14244445153153a33e7c0dd"""
        mimetypes = QtWidgets.QTreeWidget.mimeTypes(self)
        mimetypes.append(MyQTreeWidget.customMimeType)
        return mimetypes

    def startDrag(self, supportedActions):
        drag = QtGui.QDrag(self)
        mimedata = self.model().mimeData(self.selectedIndexes())

        encoded = QtCore.QByteArray()
        stream = QtCore.QDataStream(encoded, QtCore.QIODevice.WriteOnly)
        self.encodeData(self.selectedItems(), stream)
        mimedata.setData(MyQTreeWidget.customMimeType, encoded)

        drag.setMimeData(mimedata)
        drag.exec_(supportedActions)

    def encodeData(self, items: List[QtWidgets.QTreeWidgetItem],
                   stream: QtCore.QDataStream):
        stream.writeInt32(len(items))
        for item in items:
            p = item
            rows = []
            while p is not None:
                rows.append(self.indexFromItem(p).row())
                p = p.parent()
            stream.writeInt32(len(rows))
            for row in reversed(rows):
                stream.writeInt32(row)
        return stream

    def decodeData(self, encoded: QtCore.QByteArray,
                   tree: QtWidgets.QTreeWidget) -> List[QtWidgets.QTreeWidgetItem]:
        items = []
        rows = []
        stream = QtCore.QDataStream(encoded, QtCore.QIODevice.ReadOnly)
        while not stream.atEnd():
            nItems = stream.readInt32()
            for _ in range(nItems):
                path = stream.readInt32()
                row = []
                for _ in range(path):
                    row.append(stream.readInt32())
                rows.append(row)

        for row in rows:
            it = tree.topLevelItem(row[0])
            for ix in row[1:]:
                it = it.child(ix)
            items.append(it)
        return items

    def dropEvent(self, event):
        """ Reorders the recipe if event comes from self.
        Moves a step from another recipe to self if event is of type MyQTreeWidget.
        Else event from controlcenter, adds step for controlcenter """
        self.setGraphicsEffect(None)

        if event.source() is self:  # if event comes frop recipe -> reorder
            self.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)  # needed to not create new step but reorder
            QtWidgets.QTreeWidget.dropEvent(self, event)
            self.setDragDropMode(QtWidgets.QAbstractItemView.DragDrop)
            self.reorderSignal.emit(event)
        elif isinstance(event.source(), MyQTreeWidget):  # if event comes from another recipe -> remove from incoming recipe and add to outgoing recipe
            if event.mimeData().hasFormat(MyQTreeWidget.customMimeType):
                encoded = event.mimeData().data(MyQTreeWidget.customMimeType)
                items = self.decodeData(encoded, event.source())
                recipe_name_output = event.source().recipe_name

                for it in items:  # should always have one item
                    name = it.text(0)
                    stepType = self.scanner.configManager.getRecipeStepType(recipe_name_output, name)
                    stepElement = self.scanner.configManager.getRecipeStepElement(recipe_name_output, name)
                    stepValue = self.scanner.configManager.getRecipeStepValue(recipe_name_output, name)
                    # OPTIMIZE: prevent cycle between recipe and twice same recipe in a recipe before accepting the drag event

                    if stepElement != self.recipe_name:  # should always be different thanks to drag event refusal of dropping recipe in itself
                        self.scanner.configManager.configHistory.active = False
                        self.scanner.configManager.delRecipeStep(recipe_name_output, name)
                        self.scanner.configManager.configHistory.active = True
                        self.scanner.configManager.addRecipeStep(self.recipe_name, stepType, stepElement, name, stepValue)

                        try: self.scanner.configManager.getAllowedRecipe(self.recipe_name)  # OBSOLETE
                        except ValueError:
                            self.scanner.setStatus('ERROR cycle between recipes detected! change cancelled', 10000, False)
                            self.scanner.configManager.configHistory.active = False
                            self.scanner.configManager.configHistory.go_down()  # to overwrite config with error
                            self.scanner.configManager.delRecipeStep(self.recipe_name, name)
                            self.scanner.configManager.configHistory.active = True
                            self.scanner.configManager.addRecipeStep(recipe_name_output, stepType, stepElement, name, stepValue)

        else: # event comes from controlcenter -> check type to add step
            if hasattr(event.source(), 'gui'): gui = event.source().gui
            else: # OPTIMIZE: event.source() should be either self or instance of MyQTreeWidget but is sometime not recognize as such so goes to else and raise error don't have gui
                message = 'ERROR catched, please create an issue here: https://github.com/autolab-project/autolab/issues with the following informations:'
                message += '\nTitle: event.source() error in scanner'
                message += "\nDescription:"
                message += f"\nEvent source: {event.source()}"
                if hasattr(event.source(), 'customMimeType'):
                    message += f"\nEvent Mime: {event.source().customMimeType}"
                message += f"\nIs MyQTreeWidget: {isinstance(event.source(), MyQTreeWidget)}"
                message += f"\nself: {self}"
                message += f"\nIs self: {event.source() is self}"
                self.scanner.setStatus(message, 0, False)
                return None

            variable = event.source().last_drag

            if variable:
                if variable._element_type == "variable":
                    if variable.readable and variable.writable:
                        self.menu(gui, variable, event.pos())
                    elif variable.readable:
                        gui.addStepToScanRecipe(self.recipe_name, 'measure', variable)
                    elif variable.writable:
                        value = variable.value if variable.type in [tuple] else None
                        gui.addStepToScanRecipe(
                            self.recipe_name, 'set', variable, value=value)
                elif variable._element_type == "action":
                    gui.addStepToScanRecipe(self.recipe_name, 'action', variable)

            # OPTIMIZE: should be done in control panel somehow to catch drop cancel
            self.scanner.mainGui.setWindowState(
                self.scanner.mainGui.windowState() & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
            self.scanner.mainGui.activateWindow()

    def dragEnterEvent(self, event):
        """ Check if drag event is acceptable """
        # Could use mimeData instead of last_drag but overkill
        if (event.source() is self) or (
                hasattr(event.source(), "last_drag") and hasattr(event.source().last_drag, "_element_type") and event.source().last_drag._element_type != "module"):
            event.accept()
            shadow = QtWidgets.QGraphicsDropShadowEffect(blurRadius=25, xOffset=3, yOffset=3)
            self.setGraphicsEffect(shadow)

        elif isinstance(event.source(), type(self)):
            try:  # Refuse drop recipe in itself
                if event.mimeData().hasFormat(MyQTreeWidget.customMimeType):
                    encoded = event.mimeData().data(MyQTreeWidget.customMimeType)
                    items = self.decodeData(encoded, event.source())
                    recipe_name_output = event.source().recipe_name
                    for it in items:
                        name = it.text(0)
                        stepElement = self.scanner.configManager.getRecipeStepElement(
                            recipe_name_output, name)
                        if self.recipe_name == stepElement:  # OPTIMIZE: not enought to prevent cycle (see how to use config.getAllowedRecipe)
                            event.ignore()
                            return None
            except: pass

            shadow = QtWidgets.QGraphicsDropShadowEffect(blurRadius=25, xOffset=3, yOffset=3)
            event.source().setGraphicsEffect(None)
            self.setGraphicsEffect(shadow)
            event.accept()
        else:
            event.ignore()

    def dragLeaveEvent(self, event):
        self.setGraphicsEffect(None)

    def menu(self, gui: QtWidgets.QMainWindow,
             variable: Union[Variable_og, Action], position: QtCore.QPoint):
        """ Provides the menu when the user right click on an item """
        menu = QtWidgets.QMenu()
        scanMeasureStepAction = menu.addAction("Measure in scan recipe")
        scanMeasureStepAction.setIcon(icons['measure'])
        scanSetStepAction = menu.addAction("Set value in scan recipe")
        scanSetStepAction.setIcon(icons['write'])
        scanMeasureStepAction.setEnabled(variable.readable)
        scanSetStepAction.setEnabled(variable.writable)
        choice = menu.exec_(self.viewport().mapToGlobal(position))
        if choice == scanMeasureStepAction:
            gui.addStepToScanRecipe(self.recipe_name, 'measure', variable)

        elif choice == scanSetStepAction:
            value = variable.value if variable.type in [tuple] else None
            gui.addStepToScanRecipe(
                self.recipe_name, 'set', variable, value=value)

    def keyPressEvent(self, event):
        ctrl = QtCore.Qt.ControlModifier
        shift = QtCore.Qt.ShiftModifier
        mod = event.modifiers()
        if event.key() == QtCore.Qt.Key_R and mod == ctrl:
            self.rename_step(event)
        elif event.key() == QtCore.Qt.Key_C and mod == ctrl:
            self.copy_step(event)
        elif event.key() == QtCore.Qt.Key_V and mod == ctrl:
            self.paste_step(event)
        elif event.key() == QtCore.Qt.Key_Z and mod == ctrl:
            # Note: needed to add setFocus to tree on creation to allow multiple ctrl+z
            self.scanner.configManager.undoClicked()
        elif (
            event.key() == QtCore.Qt.Key_Z and mod == (ctrl | shift)
        ) or (
            event.key() == QtCore.Qt.Key_Y and mod == ctrl
        ):
            self.scanner.configManager.redoClicked()
        elif (event.key() == QtCore.Qt.Key_Delete):
            self.remove_step(event)
        else:
            super().keyPressEvent(event)

    def rename_step(self, event):
        if len(self.selectedItems()) == 0:
            super().keyPressEvent(event)
            return None
        item = self.selectedItems()[0]  # assume can select only one item
        self.scanner.recipeDict[self.recipe_name]['recipeManager'].renameStep(item.text(0))

    def copy_step(self, event):
        if len(self.selectedItems()) == 0:
            super().keyPressEvent(event)
            return None
        item = self.selectedItems()[0]  # assume can select only one item
        self.scanner.recipeDict[self.recipe_name]['recipeManager'].copyStep(item.text(0))

    def paste_step(self, event):
        self.scanner.recipeDict[self.recipe_name]['recipeManager'].pasteStep()

    def remove_step(self, event):
        if len(self.selectedItems()) == 0:
            super().keyPressEvent(event)
            return None
        item = self.selectedItems()[0]  # assume can select only one item
        self.scanner.recipeDict[self.recipe_name]['recipeManager'].removeStep(item.text(0))


class MyQTreeWidgetItem(QtWidgets.QTreeWidgetItem):

    def __init__(self, itemParent: QtWidgets.QTreeWidget, step: dict,
                 gui: QtWidgets.QMainWindow):

        self.gui = gui

        # Import Autolab config
        scanner_config = get_scanner_config()
        self.precision = scanner_config['precision']

        super().__init__(itemParent)

        self.setFlags(self.flags() ^ QtCore.Qt.ItemIsDropEnabled)
        self.setToolTip(0, step['element']._help)

        # Column 1 : Step name
        self.setText(0, step['name'])

        # OPTIMIZE: stepType is a bad name. Possible confusion with element type. stepType should be stepAction or just action
        # Column 2 : Step type
        if step['stepType'] == 'measure':
            self.setText(1, 'Measure')
            self.setIcon(0, icons['measure'])
        elif step['stepType']  == 'set':
            self.setText(1, 'Set')
            self.setIcon(0, icons['write'])
        elif step['stepType']  == 'action':
            self.setText(1, 'Do')
            self.setIcon(0, icons['action'])
        elif step['stepType']  == 'recipe':
            self.setText(1, 'Recipe')
            self.setIcon(0, icons['recipe'])

        # Column 3 : Element address
        if step['stepType'] == 'recipe':
            self.setText(2, step['element'])
        else:
            self.setText(2, step['element'].address())

        # Column 4 : Icon of element type
        etype = step['element'].type
        if etype is int: self.setIcon(3, icons['int'])
        elif etype is float: self.setIcon(3, icons['float'])
        elif etype is bool: self.setIcon(3, icons['bool'])
        elif etype is str: self.setIcon(3, icons['str'])
        elif etype is bytes: self.setIcon(3, icons['bytes'])
        elif etype is tuple: self.setIcon(3, icons['tuple'])
        elif etype is np.ndarray: self.setIcon(3, icons['ndarray'])
        elif etype is pd.DataFrame: self.setIcon(3, icons['DataFrame'])

        # Column 5 : Value if stepType is 'set'
        value = step['value']
        if value is not None:
            if has_eval(value):
                self.setText(4, f'{value}')
            else:
                try:
                    if step['element'].type in [bool, str, tuple]:
                        self.setText(4, f'{value}')
                    elif step['element'].type in [bytes]:
                        self.setText(4, f"{value.decode()}")
                    elif step['element'].type in [np.ndarray]:
                        value = array_to_str(
                            value, threshold=1000000, max_line_width=100)
                        self.setText(4, f'{value}')
                    elif step['element'].type in [pd.DataFrame]:
                        value = dataframe_to_str(value, threshold=1000000)
                        self.setText(4, f'{value}')
                    else:
                       self.setText(4, f'{value:.{self.precision}g}')
                except ValueError:
                    self.setText(4, f'{value}')

        # Column 6 : Unit of element
        unit = step['element'].unit
        if unit is not None:
            self.setText(5, str(unit))

        # set AlignTop to all columns
        for i in range(self.columnCount()):
            self.setTextAlignment(i, QtCore.Qt.AlignTop)
            # OPTIMIZE: icon are not aligned with text: https://www.xingyulei.com/post/qt-button-alignment/index.html


class MyQTabWidget(QtWidgets.QTabWidget):

    def __init__(self, frame:  QtWidgets.QFrame,
                 gui: QtWidgets.QMainWindow, recipe_name: str):

        self.recipe_name = recipe_name
        self.gui = gui
        super().__init__()

        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.menu)
        self.tabBarDoubleClicked.connect(self.renameRecipe)

        self.addTab(frame, self.recipe_name)
        stylesheet = """
            QTabWidget>QWidget>QWidget{background: rgb(240, 240, 240);}
            """
        self.setStyleSheet(stylesheet)

    def menu(self, position: QtCore.QPoint):
        """ Provides the menu when the user right click on an item """
        if not self.gui.scanManager.isStarted():

            menu = QtWidgets.QMenu()

            IS_ACTIVE = self.gui.configManager.getActive(self.recipe_name)

            if IS_ACTIVE:
                activateRecipeAction = menu.addAction("Disable recipe")
                activateRecipeAction.setIcon(icons['is-enable'])
            else:
                activateRecipeAction = menu.addAction("Enable recipe")
                activateRecipeAction.setIcon(icons['is-disable'])

            menu.addSeparator()

            renameRecipeAction = menu.addAction("Rename recipe")
            renameRecipeAction.setIcon(icons['rename'])
            removeRecipeAction = menu.addAction("Remove recipe")
            removeRecipeAction.setIcon(icons['remove'])

            menu.addSeparator()

            # OBSOLETE
            recipeLink = self.gui.configManager.getRecipeLink(self.recipe_name)
            if len(recipeLink) == 1:  # A bit too restrictive but do the work
                renameRecipeAction.setEnabled(True)
            else:
                renameRecipeAction.setEnabled(False)

            addParameterAction = menu.addAction("Add parameter")
            addParameterAction.setIcon(icons['add'])

            menu.addSeparator()

            moveUpRecipeAction = menu.addAction("Move recipe up")
            moveUpRecipeAction.setIcon(icons['up'])
            moveDownRecipeAction = menu.addAction("Move recipe down")
            moveDownRecipeAction.setIcon(icons['down'])

            config = self.gui.configManager.config
            keys = list(config)
            pos = keys.index(self.recipe_name)

            if pos == 0: moveUpRecipeAction.setEnabled(False)
            if pos == (len(keys) - 1): moveDownRecipeAction.setEnabled(False)

            choice = menu.exec_(self.mapToGlobal(QtCore.QPoint(0, 20)))  # don't use position to avoid confusion with tree element when very closed

            if choice == renameRecipeAction:
                self.renameRecipe()
            elif choice == removeRecipeAction:
                self.gui.configManager.removeRecipe(self.recipe_name)
            elif choice == activateRecipeAction:
                self.gui.configManager.activateRecipe(self.recipe_name, not IS_ACTIVE)
            elif choice == moveUpRecipeAction:
                keys[pos], keys[pos-1] = keys[pos-1], keys[pos]
                self.gui.configManager.setRecipeOrder(keys)
            elif choice == moveDownRecipeAction:
                keys[pos], keys[pos+1] = keys[pos+1], keys[pos]
                self.gui.configManager.setRecipeOrder(keys)
            elif choice == addParameterAction:
                self.gui.configManager.addParameter(self.recipe_name)

    def renameRecipe(self):
        """ Prompts the user for a new recipe name and apply it to the selected recipe """
        if not self.gui.scanManager.isStarted():
            newName, state = QtWidgets.QInputDialog.getText(
                self.gui, self.recipe_name, f"Set {self.recipe_name} new name",
                QtWidgets.QLineEdit.Normal, self.recipe_name)

            newName = clean_string(newName)

            if newName != '':
                self.gui.configManager.renameRecipe(
                    self.recipe_name, newName)


class ParameterQFrame(QtWidgets.QFrame):
    # customMimeType = "autolab/MyQTreeWidget-selectedItems"

    def __init__(self, parent: QtWidgets.QMainWindow, recipe_name: str, param_name: str):
        self.recipe_name = recipe_name
        self.param_name = param_name
        super().__init__(parent)
        self.setAcceptDrops(True)

    def dropEvent(self, event):
        """ Sets parameter if event comes from controlcenter """
        if hasattr(event.source(), "last_drag"):
            gui = event.source().gui
            variable = event.source().last_drag
            if variable and variable.parameter_allowed:
                gui.setScanParameter(self.recipe_name, self.param_name, variable)

        self.setGraphicsEffect(None)

    def dragEnterEvent(self, event):
        """ Only accept parameter from controlcenter """
        # OPTIMIZE: create mimedata like for recipe if want to drag/drop parameter to recipe or parap to param
        if (hasattr(event.source(), "last_drag") and (hasattr(event.source().last_drag, "parameter_allowed") and event.source().last_drag.parameter_allowed)):
            event.accept()
            shadow = QtWidgets.QGraphicsDropShadowEffect(blurRadius=25, xOffset=3, yOffset=3)
            self.setGraphicsEffect(shadow)
        else:
            event.ignore()

    def dragLeaveEvent(self, event):
        self.setGraphicsEffect(None)



# -*- coding: utf-8 -*-
"""
Created on Wed Apr 17 09:27:19 2024

@author: Jonathan
"""


class Driver_parser():
    def __init__(self, Instance, name, **kwargs):
        self.name     = name
        self.Instance = Instance


    def add_parser_usage(self,message):
        """Usage to be used by the parser"""
        usage = f"""
{message}

----------------  Examples:  ----------------

usage:    autolab driver [options] args

    autolab driver -D {self.name} -A GPIB::2::INSTR -C VISA
    load {self.name} driver using VISA communication protocol with address GPIB...

    autolab driver -D nickname
    same as before but using the device nickname as defined in local_config.ini
            """
        return usage

    def add_parser_arguments(self,parser):
        """Add arguments to the parser passed as input"""
        # parser.add_argument("-p", "--power", type=str, dest="power", default=None, help="Set the pump power setpoint in 1/10dBm (e.g. 100 <=> 10dBm; this conversion may depend on the model, please verify)." )

        return parser

    def do_something(self,args):
        pass
        # if args.power:
        #     getattr(self.Instance,'set_power')(args.power)

    def exit(self):
        self.Instance.close()



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 29 18:26:32 2019

@author: qchat
"""

import sys
import inspect
from typing import Any

from qtpy import QtCore, QtWidgets

from ..GUI_utilities import qt_object_exists
from ...devices import get_final_device_config, Device
from ...drivers import load_driver_lib, get_driver


class ThreadManager:
    """ This class is dedicated to manage the different threads,
    from their creation, to their deletion after they have been used """

    def __init__(self, gui: QtWidgets.QMainWindow):
        self.gui = gui
        self.threads = {}

    def start(self, item: QtWidgets.QTreeWidgetItem, intType: str, value = None):
        """ This function is called when a new thread is requested,
        for a particular intType interaction type """
        # GUI disabling
        item.setDisabled(True)

        if hasattr(item, "execButton"):
            item.execButton.setEnabled(False)
        if hasattr(item, "readButton"):
            item.readButton.setEnabled(False)
        if hasattr(item, "valueWidget"):
            item.valueWidget.setEnabled(False)

        # disabling valueWidget deselect item and select next one, need to disable all items and reenable item
        for item_selected in self.gui.tree.selectedItems():
                item_selected.setSelected(False)
        item.setSelected(True)

        # Status writing
        if intType == 'read': status = f'Reading {item.variable.address()}...'
        elif intType == 'write': status = f'Writing {item.variable.address()}...'
        elif intType == 'execute': status = f'Executing {item.action.address()}...'
        elif intType == 'load': status = f'Loading plugin {item.name}...'
        self.gui.setStatus(status)

        # Thread configuration
        thread = InteractionThread(item,intType,value)
        tid = id(thread)
        self.threads[tid] = thread
        thread.endSignal.connect(lambda error, x=tid: self.threadFinished(x, error))
        thread.finished.connect(lambda x=tid: self.delete(x))

        # Starting thread
        thread.start()

    def threadFinished(self, tid: int, error: Exception):
        """ This function is called when a thread has finished its job, with an error or not
        It updates the status bar of the GUI in consequence and enabled back the correspondig item """
        if error:
            if qt_object_exists(self.gui.statusBar):
                self.gui.setStatus(str(error), 10000, False)
            else:
                print(str(error), file=sys.stderr)
        else:
            if qt_object_exists(self.gui.statusBar):
                self.gui.clearStatus()

        item = self.threads[tid].item

        item.setDisabled(False)

        if hasattr(item, "execButton") and qt_object_exists(item.execButton):
            item.execButton.setEnabled(True)
        if hasattr(item, "readButton") and qt_object_exists(item.readButton):
            item.readButton.setEnabled(True)
        if hasattr(item, "valueWidget") and qt_object_exists(item.valueWidget):
            item.valueWidget.setEnabled(True)
            # Put back focus if item still selected (item.isSelected() doesn't work)
            if item in self.gui.tree.selectedItems():
                item.valueWidget.setFocus()

    def delete(self, tid: int):
        """ This function is called when a thread is about to be deleted.
        This removes it from the dictionnary self.threads, for a complete deletion """
        self.threads.pop(tid)


class InteractionThread(QtCore.QThread):
    """ This class is dedicated to operation interaction with the devices, in a new thread """
    endSignal = QtCore.Signal(object)

    def __init__(self, item: QtWidgets.QTreeWidgetItem, intType: str, value: Any):
        super().__init__()
        self.item = item
        self.intType = intType
        self.value = value

    def run(self):
        """ Depending on the interaction type requested, this function reads or
        writes a variable, or execute an action. """
        error = None
        try:
            if self.intType == 'read': self.item.variable()
            elif self.intType == 'write':
                self.item.variable(self.value)
                if self.item.variable.readable: self.item.variable()
            elif self.intType == 'execute':
                if self.value is not None:
                    self.item.action(self.value)
                else:
                    self.item.action()
            elif self.intType == 'load':
                # Note that threadItemDict needs to be updated outside of thread to avoid timing error
                plugin_name = self.item.name
                device_config = get_final_device_config(plugin_name)

                driver_kwargs = {k: v for k, v in device_config.items()
                                 if k not in ['driver', 'connection']}
                driver_lib = load_driver_lib(device_config['driver'])

                if hasattr(driver_lib, 'Driver') and 'gui' in [
                        param.name for param in inspect.signature(
                            driver_lib.Driver.__init__).parameters.values()]:

                        driver_kwargs['gui'] = self.item.gui

                instance = get_driver(
                    device_config['driver'], device_config['connection'],
                    **driver_kwargs)
                module = Device(plugin_name, instance, device_config)
                self.item.gui.threadDeviceDict[id(self.item)] = module
        except Exception as e:
            error = e
            if self.intType == 'load':
                error = f'An error occured when loading device {self.item.name}: {str(e)}'
                if id(self.item) in self.item.gui.threadItemDict.keys():
                    self.item.gui.threadItemDict.pop(id(self.item))
        self.endSignal.emit(error)
